#include <sivkernel/interface.h>
#include <sivkernel/method.h>
#include <sivkernel/kernel.h>
#include <sivkernel/generated/library.h>
#include <sivdk/exception_kind.h>
#include <sivkernel-mocks/domain_mock.h>
#include <sivdk/interface_handle.h>
#include <sivdk/value_type_handle.h>
#include <sivdk/exception_kind.h>
#include <sivtypes/messengers/device_type.h>
#include <gtest/gtest.h>
#include <gmock/gmock.h>
#include <string>
#include <unordered_set>

using namespace sivelkiria::kernel;
using namespace sivelkiria::kernel::generated;
using namespace sivtypes::messengers;
using namespace mocks;
using namespace sivelkiria::common;
using namespace data_pack;
using namespace std;
using namespace testing;


TEST(DynamicStructures, ReturnValue)
{
    auto *const library = library::get_instance();

    kernel kernel(library);
    auto domain1 = make_shared<local_domain_mock>(&kernel);

    const auto *const body = library->get_interface(interface_handle::messengers__contact);
    const method_base *get_status = nullptr;
    const method_id get_status_id = 0;
    ASSERT_EQ(body->get_supported_method(get_status_id, 0, get_status), method_lookup_result::implemented);

    auto host1 = domain1->register_module_side_host(body->get_handle(), 0);

    handle_type pack[100];
    object_handle releases[100];
    uint64_t release_counts[100];
    void* release_pointers[100];
    uint64_t release_count = 0;

    {
        handle_ptr<outbound_invocation<inbound_local_method_data>, module_to_kernel_invocation_handle> handle(
            domain1->get_outbound_invocations());
        auto* const request = static_cast<inbound_local_method_data*>(handle->get_inbound_method_data());
        request->set_request(host1, body->get_handle(), get_status_id, pack, releases, release_counts,
                             reinterpret_cast<char*>(release_pointers), &release_count);

        ASSERT_TRUE(domain1->start_invocation(handle.get_handle()));
        ASSERT_TRUE(handle);
        EXPECT_CALL(*domain1, send_request(_))
            .Times(1)
            .WillOnce(Return(domain_base::request_sending_status::pending_response));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());
        auto* local_data_passed = dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data());
        ASSERT_NE(local_data_passed, nullptr);

        object_handle target_object;
        interface_handle target_interface;
        method_id target_method;
        dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data())
            ->get_request_method_info(target_object, target_interface, target_method);
        ASSERT_EQ(target_object, host1);
        ASSERT_EQ(target_interface, body->get_handle());
        ASSERT_EQ(target_method, method_id(get_status_id));

        local_data_passed->access_return_value()
            .write(value_type_handle::messengers__status)
            .write(nullptr);
        local_data_passed->write_exception_data(exception_handle::invalid);

        auto callee_side_handle = handle->get_callee_side_handle();
        ASSERT_TRUE(domain1->process_response(callee_side_handle));
        EXPECT_CALL(*domain1,
                    pass_response_data_to_module(domain1->get_outbound_invocations()->access(handle.get_handle())))
            .Times(1)
            .WillOnce(Return(response_sending_status::wait_for_modules_confirmation));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());

        exception_handle error_handle = static_cast<exception_handle>(static_cast<handle_type>(pack[0]));
        exception_kind error_kind = static_cast<exception_kind>(static_cast<handle_type>(pack[1]));
        ASSERT_EQ(error_handle, exception_handle::invalid);
        ASSERT_EQ(error_kind, static_cast<exception_kind>(static_cast<uint64_t>(invalid_interface_handle)));

        value_type_handle type = value_type_handle::invalid;
        void* data = nullptr;
        pack_field_const_iterator(&pack[2], endian::native)
            .read(&type)
            .read(&data);

        EXPECT_EQ(type, value_type_handle::messengers__status);
        EXPECT_EQ(data, nullptr);
    }

    {
        handle_ptr<outbound_invocation<inbound_local_method_data>, module_to_kernel_invocation_handle> handle(
            domain1->get_outbound_invocations());
        auto* const request = static_cast<inbound_local_method_data*>(handle->get_inbound_method_data());
        request->set_request(host1, body->get_handle(), get_status_id, pack, releases, release_counts,
                             reinterpret_cast<char*>(release_pointers), &release_count);

        ASSERT_TRUE(domain1->start_invocation(handle.get_handle()));
        ASSERT_TRUE(handle);
        EXPECT_CALL(*domain1, send_request(_))
            .Times(1)
            .WillOnce(Return(domain_base::request_sending_status::pending_response));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());
        auto* local_data_passed = dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data());
        ASSERT_NE(local_data_passed, nullptr);

        object_handle target_object;
        interface_handle target_interface;
        method_id target_method;
        dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data())
            ->get_request_method_info(target_object, target_interface, target_method);
        ASSERT_EQ(target_object, host1);
        ASSERT_EQ(target_interface, body->get_handle());
        ASSERT_EQ(target_method, method_id(get_status_id));

        struct offline_body {
            alignas(handle_type) uint64_t m_minutes;
        } sent {
            50
        };

        local_data_passed->access_return_value()
            .write(value_type_handle::messengers__status_offline)
            .write(&sent);
        local_data_passed->write_exception_data(exception_handle::invalid);

        auto callee_side_handle = handle->get_callee_side_handle();
        ASSERT_TRUE(domain1->process_response(callee_side_handle));
        EXPECT_CALL(*domain1,
                    pass_response_data_to_module(domain1->get_outbound_invocations()->access(handle.get_handle())))
            .Times(1)
            .WillOnce(Return(response_sending_status::wait_for_modules_confirmation));
        EXPECT_CALL(*domain1, request_module_free_response_dynamic_data(callee_side_handle, _)).Times(1);
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());

        exception_handle error_handle = static_cast<exception_handle>(static_cast<handle_type>(pack[0]));
        exception_kind error_kind = static_cast<exception_kind>(static_cast<handle_type>(pack[1]));
        ASSERT_EQ(error_handle, exception_handle::invalid);
        ASSERT_EQ(error_kind, static_cast<exception_kind>(static_cast<uint64_t>(invalid_interface_handle)));

        value_type_handle type = value_type_handle::invalid;
        offline_body* received = nullptr;
        pack_field_const_iterator(&pack[2], endian::native)
            .read(&type)
            .read(&received);

        EXPECT_EQ(type, value_type_handle::messengers__status_offline);
        EXPECT_NE(received, nullptr);
        EXPECT_NE(received, &sent);
        EXPECT_EQ(received->m_minutes, 50);
    }

    {
        handle_ptr<outbound_invocation<inbound_local_method_data>, module_to_kernel_invocation_handle> handle(
            domain1->get_outbound_invocations());
        auto* const request = static_cast<inbound_local_method_data*>(handle->get_inbound_method_data());
        request->set_request(host1, body->get_handle(), get_status_id, pack, releases, release_counts,
                             reinterpret_cast<char*>(release_pointers), &release_count);

        ASSERT_TRUE(domain1->start_invocation(handle.get_handle()));
        ASSERT_TRUE(handle);
        EXPECT_CALL(*domain1, send_request(_))
            .Times(1)
            .WillOnce(Return(domain_base::request_sending_status::pending_response));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());
        auto* local_data_passed = dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data());
        ASSERT_NE(local_data_passed, nullptr);

        object_handle target_object;
        interface_handle target_interface;
        method_id target_method;
        dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data())
            ->get_request_method_info(target_object, target_interface, target_method);
        ASSERT_EQ(target_object, host1);
        ASSERT_EQ(target_interface, body->get_handle());
        ASSERT_EQ(target_method, method_id(get_status_id));

        struct away_body {
            alignas(handle_type) device_type m_device_type;
            alignas(handle_type) uint64_t m_minutes;
        } sent {
            device_type::mobile_device,
            15
        };

        local_data_passed->access_return_value()
            .write(value_type_handle::messengers__status_online_away)
            .write(&sent);
        local_data_passed->write_exception_data(exception_handle::invalid);

        auto callee_side_handle = handle->get_callee_side_handle();
        ASSERT_TRUE(domain1->process_response(callee_side_handle));
        EXPECT_CALL(*domain1,
                    pass_response_data_to_module(domain1->get_outbound_invocations()->access(handle.get_handle())))
            .Times(1)
            .WillOnce(Return(response_sending_status::wait_for_modules_confirmation));
        EXPECT_CALL(*domain1, request_module_free_response_dynamic_data(callee_side_handle, _)).Times(1);
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());

        exception_handle error_handle = static_cast<exception_handle>(static_cast<handle_type>(pack[0]));
        exception_kind error_kind = static_cast<exception_kind>(static_cast<handle_type>(pack[1]));
        ASSERT_EQ(error_handle, exception_handle::invalid);
        ASSERT_EQ(error_kind, static_cast<exception_kind>(static_cast<uint64_t>(invalid_interface_handle)));

        value_type_handle type = value_type_handle::invalid;
        away_body* received = nullptr;
        pack_field_const_iterator(&pack[2], endian::native)
            .read(&type)
            .read(&received);

        EXPECT_EQ(type, value_type_handle::messengers__status_online_away);
        EXPECT_NE(received, nullptr);
        EXPECT_NE(received, &sent);
        EXPECT_EQ(received->m_device_type, device_type::mobile_device);
        EXPECT_EQ(received->m_minutes, 15);
    }

    {
        handle_ptr<outbound_invocation<inbound_local_method_data>, module_to_kernel_invocation_handle> handle(
            domain1->get_outbound_invocations());
        auto* const request = static_cast<inbound_local_method_data*>(handle->get_inbound_method_data());
        request->set_request(host1, body->get_handle(), get_status_id, pack, releases, release_counts,
                             reinterpret_cast<char*>(release_pointers), &release_count);

        ASSERT_TRUE(domain1->start_invocation(handle.get_handle()));
        ASSERT_TRUE(handle);
        EXPECT_CALL(*domain1, send_request(_))
            .Times(1)
            .WillOnce(Return(domain_base::request_sending_status::pending_response));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());
        auto* local_data_passed = dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data());
        ASSERT_NE(local_data_passed, nullptr);

        object_handle target_object;
        interface_handle target_interface;
        method_id target_method;
        dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data())
            ->get_request_method_info(target_object, target_interface, target_method);
        ASSERT_EQ(target_object, host1);
        ASSERT_EQ(target_interface, body->get_handle());
        ASSERT_EQ(target_method, method_id(get_status_id));

        local_data_passed->access_return_value()
            .write(value_type_handle::invalid)
            .write(nullptr);
        local_data_passed->write_exception_data(exception_handle::invalid);

        auto callee_side_handle = handle->get_callee_side_handle();
        ASSERT_TRUE(domain1->process_response(callee_side_handle));
        EXPECT_CALL(*domain1,
                    pass_response_data_to_module(domain1->get_outbound_invocations()->access(handle.get_handle())))
            .Times(1)
            .WillOnce(Return(response_sending_status::wait_for_modules_confirmation));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());

        exception_handle error_handle = static_cast<exception_handle>(static_cast<handle_type>(pack[0]));
        exception_kind error_kind = static_cast<exception_kind>(static_cast<handle_type>(pack[1]));
        ASSERT_EQ(error_handle, exception_handle::invalid);
        ASSERT_EQ(error_kind, static_cast<exception_kind>(static_cast<uint64_t>(invalid_interface_handle)));

        value_type_handle type = value_type_handle::invalid;
        void* data = nullptr;
        pack_field_const_iterator(&pack[2], endian::native)
            .read(&type)
            .read(&data);

        EXPECT_EQ(type, value_type_handle::invalid);
        EXPECT_EQ(data, nullptr);
    }
}

TEST(DynamicStructures, Parameter)
{
    auto *const library = library::get_instance();

    kernel kernel(library);
    auto domain1 = make_shared<local_domain_mock>(&kernel);

    const auto *const body = library->get_interface(interface_handle::messengers__status_handler);
    const method_base *truncate = nullptr;
    const method_id truncate_id = 0;
    ASSERT_EQ(body->get_supported_method(truncate_id, 0, truncate), method_lookup_result::implemented);

    auto host1 = domain1->register_module_side_host(body->get_handle(), 0);

    handle_type pack[100];
    object_handle releases[100];
    uint64_t release_counts[100];
    void* release_pointers[100];
    uint64_t release_count = 0;

    {
        pack_field_iterator(&pack[2], endian::native)
            .write(value_type_handle::messengers__status)
            .write(nullptr);

        handle_ptr<outbound_invocation<inbound_local_method_data>, module_to_kernel_invocation_handle> handle(
            domain1->get_outbound_invocations());
        auto* const request = static_cast<inbound_local_method_data*>(handle->get_inbound_method_data());
        request->set_request(host1, body->get_handle(), truncate_id, pack, releases, release_counts,
                             reinterpret_cast<char*>(release_pointers), &release_count);

        ASSERT_TRUE(domain1->start_invocation(handle.get_handle()));
        ASSERT_TRUE(handle);
        EXPECT_CALL(*domain1, send_request(_))
            .Times(1)
            .WillOnce(Return(domain_base::request_sending_status::pending_response));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());
        auto* local_data_passed = dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data());
        ASSERT_NE(local_data_passed, nullptr);

        object_handle target_object;
        interface_handle target_interface;
        method_id target_method;
        dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data())
            ->get_request_method_info(target_object, target_interface, target_method);
        ASSERT_EQ(target_object, host1);
        ASSERT_EQ(target_interface, body->get_handle());
        ASSERT_EQ(target_method, method_id(truncate_id));

        value_type_handle received_type = value_type_handle::invalid;
        void* received_data = nullptr;
        local_data_passed->access_parameters(0)
            .read(&received_type)
            .read(&received_data);
        EXPECT_EQ(received_type, value_type_handle::messengers__status);
        EXPECT_EQ(received_data, nullptr);

        local_data_passed->write_exception_data(exception_handle::invalid);

        auto callee_side_handle = handle->get_callee_side_handle();
        ASSERT_TRUE(domain1->process_response(callee_side_handle));
        EXPECT_CALL(*domain1,
                    pass_response_data_to_module(domain1->get_outbound_invocations()->access(handle.get_handle())))
            .Times(1)
            .WillOnce(Return(response_sending_status::wait_for_modules_confirmation));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());

        exception_handle error_handle = static_cast<exception_handle>(static_cast<handle_type>(pack[0]));
        exception_kind error_kind = static_cast<exception_kind>(static_cast<handle_type>(pack[1]));
        ASSERT_EQ(error_handle, exception_handle::invalid);
        ASSERT_EQ(error_kind, static_cast<exception_kind>(static_cast<uint64_t>(invalid_interface_handle)));

        value_type_handle returned_type = value_type_handle::invalid;
        void* returned_data = nullptr;
        pack_field_const_iterator(&pack[2], endian::native)
            .read(&returned_type)
            .read(&returned_data);
        EXPECT_EQ(returned_type, value_type_handle::messengers__status);
        EXPECT_EQ(returned_data, nullptr);
    }

    {
        pack_field_iterator(&pack[2], endian::native)
            .write(value_type_handle::invalid)
            .write(nullptr);

        handle_ptr<outbound_invocation<inbound_local_method_data>, module_to_kernel_invocation_handle> handle(
            domain1->get_outbound_invocations());
        auto* const request = static_cast<inbound_local_method_data*>(handle->get_inbound_method_data());
        request->set_request(host1, body->get_handle(), truncate_id, pack, releases, release_counts,
                             reinterpret_cast<char*>(release_pointers), &release_count);

        ASSERT_TRUE(domain1->start_invocation(handle.get_handle()));
        ASSERT_TRUE(handle);
        EXPECT_CALL(*domain1, send_request(_))
            .Times(1)
            .WillOnce(Return(domain_base::request_sending_status::pending_response));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());
        auto* local_data_passed = dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data());
        ASSERT_NE(local_data_passed, nullptr);

        object_handle target_object;
        interface_handle target_interface;
        method_id target_method;
        dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data())
            ->get_request_method_info(target_object, target_interface, target_method);
        ASSERT_EQ(target_object, host1);
        ASSERT_EQ(target_interface, body->get_handle());
        ASSERT_EQ(target_method, method_id(truncate_id));

        value_type_handle received_type = value_type_handle::invalid;
        void* received_data = nullptr;
        local_data_passed->access_parameters(0)
            .read(&received_type)
            .read(&received_data);
        EXPECT_EQ(received_type, value_type_handle::invalid);
        EXPECT_EQ(received_data, nullptr);

        local_data_passed->write_exception_data(exception_handle::invalid);

        auto callee_side_handle = handle->get_callee_side_handle();
        ASSERT_TRUE(domain1->process_response(callee_side_handle));
        EXPECT_CALL(*domain1,
                    pass_response_data_to_module(domain1->get_outbound_invocations()->access(handle.get_handle())))
            .Times(1)
            .WillOnce(Return(response_sending_status::wait_for_modules_confirmation));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());

        exception_handle error_handle = static_cast<exception_handle>(static_cast<handle_type>(pack[0]));
        exception_kind error_kind = static_cast<exception_kind>(static_cast<handle_type>(pack[1]));
        ASSERT_EQ(error_handle, exception_handle::invalid);
        ASSERT_EQ(error_kind, static_cast<exception_kind>(static_cast<uint64_t>(invalid_interface_handle)));

        value_type_handle returned_type = value_type_handle::invalid;
        void* returned_data = nullptr;
        pack_field_const_iterator(&pack[2], endian::native)
            .read(&returned_type)
            .read(&returned_data);
        EXPECT_EQ(returned_type, value_type_handle::invalid);
        EXPECT_EQ(returned_data, nullptr);
    }

    {
        struct offline_body {
            alignas(handle_type) uint64_t m_minutes;
        } sent {
            50
        };

        pack_field_iterator(&pack[2], endian::native)
            .write(value_type_handle::messengers__status_offline)
            .write(&sent);

        handle_ptr<outbound_invocation<inbound_local_method_data>, module_to_kernel_invocation_handle> handle(
            domain1->get_outbound_invocations());
        auto* const request = static_cast<inbound_local_method_data*>(handle->get_inbound_method_data());
        request->set_request(host1, body->get_handle(), truncate_id, pack, releases, release_counts,
                             reinterpret_cast<char*>(release_pointers), &release_count);

        ASSERT_TRUE(domain1->start_invocation(handle.get_handle()));
        ASSERT_TRUE(handle);
        EXPECT_CALL(*domain1, send_request(_))
            .Times(1)
            .WillOnce(Return(domain_base::request_sending_status::pending_response));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());
        auto* local_data_passed = dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data());
        ASSERT_NE(local_data_passed, nullptr);

        object_handle target_object;
        interface_handle target_interface;
        method_id target_method;
        dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data())
            ->get_request_method_info(target_object, target_interface, target_method);
        ASSERT_EQ(target_object, host1);
        ASSERT_EQ(target_interface, body->get_handle());
        ASSERT_EQ(target_method, method_id(truncate_id));

        value_type_handle received_type = value_type_handle::invalid;
        offline_body* received_data = nullptr;
        local_data_passed->access_parameters(0)
            .read(&received_type)
            .read(&received_data);
        EXPECT_EQ(received_type, value_type_handle::messengers__status_offline);
        EXPECT_NE(received_data, nullptr);
        EXPECT_NE(received_data, &sent);
        EXPECT_EQ(received_data->m_minutes, 50);

        local_data_passed->write_exception_data(exception_handle::invalid);

        auto callee_side_handle = handle->get_callee_side_handle();
        ASSERT_TRUE(domain1->process_response(callee_side_handle));
        EXPECT_CALL(*domain1,
                    pass_response_data_to_module(domain1->get_outbound_invocations()->access(handle.get_handle())))
            .Times(1)
            .WillOnce(Return(response_sending_status::wait_for_modules_confirmation));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());

        exception_handle error_handle = static_cast<exception_handle>(static_cast<handle_type>(pack[0]));
        exception_kind error_kind = static_cast<exception_kind>(static_cast<handle_type>(pack[1]));
        ASSERT_EQ(error_handle, exception_handle::invalid);
        ASSERT_EQ(error_kind, static_cast<exception_kind>(static_cast<uint64_t>(invalid_interface_handle)));

        value_type_handle returned_type = value_type_handle::invalid;
        offline_body* returned_data = nullptr;
        pack_field_const_iterator(&pack[2], endian::native)
            .read(&returned_type)
            .read(&returned_data);
        EXPECT_EQ(returned_type, value_type_handle::messengers__status_offline);
        EXPECT_EQ(returned_data, &sent);
    }

    {
        struct away_body {
            alignas(handle_type) device_type m_device_type;
            alignas(handle_type) uint64_t m_away_minutes;
        } sent {
            device_type::bot,
            114
        };

        pack_field_iterator(&pack[2], endian::native)
            .write(value_type_handle::messengers__status_online_away)
            .write(&sent);

        handle_ptr<outbound_invocation<inbound_local_method_data>, module_to_kernel_invocation_handle> handle(
            domain1->get_outbound_invocations());
        auto* const request = static_cast<inbound_local_method_data*>(handle->get_inbound_method_data());
        request->set_request(host1, body->get_handle(), truncate_id, pack, releases, release_counts,
                             reinterpret_cast<char*>(release_pointers), &release_count);

        ASSERT_TRUE(domain1->start_invocation(handle.get_handle()));
        ASSERT_TRUE(handle);
        EXPECT_CALL(*domain1, send_request(_))
            .Times(1)
            .WillOnce(Return(domain_base::request_sending_status::pending_response));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());
        auto* local_data_passed = dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data());
        ASSERT_NE(local_data_passed, nullptr);

        object_handle target_object;
        interface_handle target_interface;
        method_id target_method;
        dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data())
            ->get_request_method_info(target_object, target_interface, target_method);
        ASSERT_EQ(target_object, host1);
        ASSERT_EQ(target_interface, body->get_handle());
        ASSERT_EQ(target_method, method_id(truncate_id));

        value_type_handle received_type = value_type_handle::invalid;
        away_body* received_data = nullptr;
        local_data_passed->access_parameters(0)
            .read(&received_type)
            .read(&received_data);
        EXPECT_EQ(received_type, value_type_handle::messengers__status_online_away);
        EXPECT_NE(received_data, nullptr);
        EXPECT_NE(received_data, &sent);
        EXPECT_EQ(received_data->m_device_type, device_type::bot);
        EXPECT_EQ(received_data->m_away_minutes, 114);

        local_data_passed->write_exception_data(exception_handle::invalid);

        auto callee_side_handle = handle->get_callee_side_handle();
        ASSERT_TRUE(domain1->process_response(callee_side_handle));
        EXPECT_CALL(*domain1,
                    pass_response_data_to_module(domain1->get_outbound_invocations()->access(handle.get_handle())))
            .Times(1)
            .WillOnce(Return(response_sending_status::wait_for_modules_confirmation));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());

        exception_handle error_handle = static_cast<exception_handle>(static_cast<handle_type>(pack[0]));
        exception_kind error_kind = static_cast<exception_kind>(static_cast<handle_type>(pack[1]));
        ASSERT_EQ(error_handle, exception_handle::invalid);
        ASSERT_EQ(error_kind, static_cast<exception_kind>(static_cast<uint64_t>(invalid_interface_handle)));

        value_type_handle returned_type = value_type_handle::invalid;
        away_body* returned_data = nullptr;
        pack_field_const_iterator(&pack[2], endian::native)
            .read(&returned_type)
            .read(&returned_data);
        EXPECT_EQ(returned_type, value_type_handle::messengers__status_online_away);
        EXPECT_EQ(returned_data, &sent);
    }
}

TEST(DynamicStructures, ParameterChanged)
{
    auto *const library = library::get_instance();

    kernel kernel(library);
    auto domain1 = make_shared<local_domain_mock>(&kernel);

    const auto *const body = library->get_interface(interface_handle::messengers__status_handler);
    const method_base *truncate = nullptr;
    const method_id truncate_id = 0;
    ASSERT_EQ(body->get_supported_method(truncate_id, 0, truncate), method_lookup_result::implemented);

    auto host1 = domain1->register_module_side_host(body->get_handle(), 0);

    handle_type pack[100];
    object_handle releases[100];
    uint64_t release_counts[100];
    void* release_pointers[100];
    uint64_t release_count = 0;

    struct offline_body {
        alignas(handle_type) uint64_t m_minutes;
    };
    struct away_body {
        alignas(handle_type) device_type m_device_type;
        alignas(handle_type) uint64_t m_away_minutes;
    };

    {
        pack_field_iterator(&pack[2], endian::native)
            .write(value_type_handle::messengers__status)
            .write(nullptr);

        handle_ptr<outbound_invocation<inbound_local_method_data>, module_to_kernel_invocation_handle> handle(
            domain1->get_outbound_invocations());
        auto* const request = static_cast<inbound_local_method_data*>(handle->get_inbound_method_data());
        request->set_request(host1, body->get_handle(), truncate_id, pack, releases, release_counts,
                             reinterpret_cast<char*>(release_pointers), &release_count);

        ASSERT_TRUE(domain1->start_invocation(handle.get_handle()));
        ASSERT_TRUE(handle);
        EXPECT_CALL(*domain1, send_request(_))
            .Times(1)
            .WillOnce(Return(domain_base::request_sending_status::pending_response));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());
        auto* local_data_passed = dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data());
        ASSERT_NE(local_data_passed, nullptr);

        object_handle target_object;
        interface_handle target_interface;
        method_id target_method;
        dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data())
            ->get_request_method_info(target_object, target_interface, target_method);
        ASSERT_EQ(target_object, host1);
        ASSERT_EQ(target_interface, body->get_handle());
        ASSERT_EQ(target_method, method_id(truncate_id));

        value_type_handle received_type = value_type_handle::invalid;
        void* received_data = nullptr;
        local_data_passed->access_parameters(0)
            .read(&received_type)
            .read(&received_data);
        EXPECT_EQ(received_type, value_type_handle::messengers__status);
        EXPECT_EQ(received_data, nullptr);

        local_data_passed->access_parameters(0)
            .write(value_type_handle::invalid)
            .write(nullptr);

        local_data_passed->write_exception_data(exception_handle::invalid);

        auto callee_side_handle = handle->get_callee_side_handle();
        ASSERT_TRUE(domain1->process_response(callee_side_handle));
        EXPECT_CALL(*domain1,
                    pass_response_data_to_module(domain1->get_outbound_invocations()->access(handle.get_handle())))
            .Times(1)
            .WillOnce(Return(response_sending_status::wait_for_modules_confirmation));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());

        exception_handle error_handle = static_cast<exception_handle>(static_cast<handle_type>(pack[0]));
        exception_kind error_kind = static_cast<exception_kind>(static_cast<handle_type>(pack[1]));
        ASSERT_EQ(error_handle, exception_handle::invalid);
        ASSERT_EQ(error_kind, static_cast<exception_kind>(static_cast<uint64_t>(invalid_interface_handle)));

        value_type_handle returned_type = value_type_handle::invalid;
        void* returned_data = nullptr;
        pack_field_const_iterator(&pack[2], endian::native)
            .read(&returned_type)
            .read(&returned_data);
        EXPECT_EQ(returned_type, value_type_handle::invalid);
        EXPECT_EQ(returned_data, nullptr);
    }

    {
        pack_field_iterator(&pack[2], endian::native)
            .write(value_type_handle::invalid)
            .write(nullptr);

        handle_ptr<outbound_invocation<inbound_local_method_data>, module_to_kernel_invocation_handle> handle(
            domain1->get_outbound_invocations());
        auto* const request = static_cast<inbound_local_method_data*>(handle->get_inbound_method_data());
        request->set_request(host1, body->get_handle(), truncate_id, pack, releases, release_counts,
                             reinterpret_cast<char*>(release_pointers), &release_count);

        ASSERT_TRUE(domain1->start_invocation(handle.get_handle()));
        ASSERT_TRUE(handle);
        EXPECT_CALL(*domain1, send_request(_))
            .Times(1)
            .WillOnce(Return(domain_base::request_sending_status::pending_response));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());
        auto* local_data_passed = dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data());
        ASSERT_NE(local_data_passed, nullptr);

        object_handle target_object;
        interface_handle target_interface;
        method_id target_method;
        dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data())
            ->get_request_method_info(target_object, target_interface, target_method);
        ASSERT_EQ(target_object, host1);
        ASSERT_EQ(target_interface, body->get_handle());
        ASSERT_EQ(target_method, method_id(truncate_id));

        value_type_handle received_type = value_type_handle::invalid;
        void* received_data = nullptr;
        local_data_passed->access_parameters(0)
            .read(&received_type)
            .read(&received_data);
        EXPECT_EQ(received_type, value_type_handle::invalid);
        EXPECT_EQ(received_data, nullptr);

        away_body changed {
            device_type::personal_computer,
            1
        };
        local_data_passed->access_parameters(0)
            .write(value_type_handle::messengers__status_online_away)
            .write(&changed);

        local_data_passed->write_exception_data(exception_handle::invalid);

        auto callee_side_handle = handle->get_callee_side_handle();
        ASSERT_TRUE(domain1->process_response(callee_side_handle));
        EXPECT_CALL(*domain1,
                    pass_response_data_to_module(domain1->get_outbound_invocations()->access(handle.get_handle())))
            .Times(1)
            .WillOnce(Return(response_sending_status::wait_for_modules_confirmation));
        EXPECT_CALL(*domain1, request_module_free_response_dynamic_data(callee_side_handle, _)).Times(1);
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());

        exception_handle error_handle = static_cast<exception_handle>(static_cast<handle_type>(pack[0]));
        exception_kind error_kind = static_cast<exception_kind>(static_cast<handle_type>(pack[1]));
        ASSERT_EQ(error_handle, exception_handle::invalid);
        ASSERT_EQ(error_kind, static_cast<exception_kind>(static_cast<uint64_t>(invalid_interface_handle)));

        value_type_handle returned_type = value_type_handle::invalid;
        away_body* returned_data = nullptr;
        pack_field_const_iterator(&pack[2], endian::native)
            .read(&returned_type)
            .read(&returned_data);
        EXPECT_EQ(returned_type, value_type_handle::messengers__status_online_away);
        EXPECT_NE(returned_data, nullptr);
        EXPECT_NE(returned_data, &changed);
        EXPECT_EQ(returned_data->m_device_type, device_type::personal_computer);
        EXPECT_EQ(returned_data->m_away_minutes, 1);
    }

    {
        offline_body sent {
            50
        };

        pack_field_iterator(&pack[2], endian::native)
            .write(value_type_handle::messengers__status_offline)
            .write(&sent);

        handle_ptr<outbound_invocation<inbound_local_method_data>, module_to_kernel_invocation_handle> handle(
            domain1->get_outbound_invocations());
        auto* const request = static_cast<inbound_local_method_data*>(handle->get_inbound_method_data());
        request->set_request(host1, body->get_handle(), truncate_id, pack, releases, release_counts,
                             reinterpret_cast<char*>(release_pointers), &release_count);

        ASSERT_TRUE(domain1->start_invocation(handle.get_handle()));
        ASSERT_TRUE(handle);
        EXPECT_CALL(*domain1, send_request(_))
            .Times(1)
            .WillOnce(Return(domain_base::request_sending_status::pending_response));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());
        auto* local_data_passed = dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data());
        ASSERT_NE(local_data_passed, nullptr);

        object_handle target_object;
        interface_handle target_interface;
        method_id target_method;
        dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data())
            ->get_request_method_info(target_object, target_interface, target_method);
        ASSERT_EQ(target_object, host1);
        ASSERT_EQ(target_interface, body->get_handle());
        ASSERT_EQ(target_method, method_id(truncate_id));

        value_type_handle received_type = value_type_handle::invalid;
        offline_body* received_data = nullptr;
        local_data_passed->access_parameters(0)
            .read(&received_type)
            .read(&received_data);
        EXPECT_EQ(received_type, value_type_handle::messengers__status_offline);
        EXPECT_NE(received_data, nullptr);
        EXPECT_NE(received_data, &sent);
        EXPECT_EQ(received_data->m_minutes, 50);

        local_data_passed->access_parameters(0)
            .write(value_type_handle::messengers__status)
            .write(nullptr);
        local_data_passed->write_exception_data(exception_handle::invalid);

        auto callee_side_handle = handle->get_callee_side_handle();
        ASSERT_TRUE(domain1->process_response(callee_side_handle));
        EXPECT_CALL(*domain1,
                    pass_response_data_to_module(domain1->get_outbound_invocations()->access(handle.get_handle())))
            .Times(1)
            .WillOnce(Return(response_sending_status::wait_for_modules_confirmation));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());

        exception_handle error_handle = static_cast<exception_handle>(static_cast<handle_type>(pack[0]));
        exception_kind error_kind = static_cast<exception_kind>(static_cast<handle_type>(pack[1]));
        ASSERT_EQ(error_handle, exception_handle::invalid);
        ASSERT_EQ(error_kind, static_cast<exception_kind>(static_cast<uint64_t>(invalid_interface_handle)));

        value_type_handle returned_type = value_type_handle::invalid;
        void* returned_data = nullptr;
        pack_field_const_iterator(&pack[2], endian::native)
            .read(&returned_type)
            .read(&returned_data);
        EXPECT_EQ(returned_type, value_type_handle::messengers__status);
        EXPECT_EQ(returned_data, nullptr);
    }

    {
        away_body sent {
            device_type::bot,
            114
        };

        pack_field_iterator(&pack[2], endian::native)
            .write(value_type_handle::messengers__status_online_away)
            .write(&sent);

        handle_ptr<outbound_invocation<inbound_local_method_data>, module_to_kernel_invocation_handle> handle(
            domain1->get_outbound_invocations());
        auto* const request = static_cast<inbound_local_method_data*>(handle->get_inbound_method_data());
        request->set_request(host1, body->get_handle(), truncate_id, pack, releases, release_counts,
                             reinterpret_cast<char*>(release_pointers), &release_count);

        ASSERT_TRUE(domain1->start_invocation(handle.get_handle()));
        ASSERT_TRUE(handle);
        EXPECT_CALL(*domain1, send_request(_))
            .Times(1)
            .WillOnce(Return(domain_base::request_sending_status::pending_response));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());
        auto* local_data_passed = dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data());
        ASSERT_NE(local_data_passed, nullptr);

        object_handle target_object;
        interface_handle target_interface;
        method_id target_method;
        dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data())
            ->get_request_method_info(target_object, target_interface, target_method);
        ASSERT_EQ(target_object, host1);
        ASSERT_EQ(target_interface, body->get_handle());
        ASSERT_EQ(target_method, method_id(truncate_id));

        value_type_handle received_type = value_type_handle::invalid;
        away_body* received_data = nullptr;
        local_data_passed->access_parameters(0)
            .read(&received_type)
            .read(&received_data);
        EXPECT_EQ(received_type, value_type_handle::messengers__status_online_away);
        EXPECT_NE(received_data, nullptr);
        EXPECT_NE(received_data, &sent);
        EXPECT_EQ(received_data->m_device_type, device_type::bot);
        EXPECT_EQ(received_data->m_away_minutes, 114);

        away_body changed {
            device_type::personal_computer,
            1
        };
        local_data_passed->access_parameters(0)
            .write(value_type_handle::messengers__status_online_away)
            .write(&changed);

        local_data_passed->write_exception_data(exception_handle::invalid);

        auto callee_side_handle = handle->get_callee_side_handle();
        ASSERT_TRUE(domain1->process_response(callee_side_handle));
        EXPECT_CALL(*domain1,
                    pass_response_data_to_module(domain1->get_outbound_invocations()->access(handle.get_handle())))
            .Times(1)
            .WillOnce(Return(response_sending_status::wait_for_modules_confirmation));
        EXPECT_CALL(*domain1, request_module_free_response_dynamic_data(callee_side_handle, _)).Times(1);
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());

        exception_handle error_handle = static_cast<exception_handle>(static_cast<handle_type>(pack[0]));
        exception_kind error_kind = static_cast<exception_kind>(static_cast<handle_type>(pack[1]));
        ASSERT_EQ(error_handle, exception_handle::invalid);
        ASSERT_EQ(error_kind, static_cast<exception_kind>(static_cast<uint64_t>(invalid_interface_handle)));

        value_type_handle returned_type = value_type_handle::invalid;
        away_body* returned_data = nullptr;
        pack_field_const_iterator(&pack[2], endian::native)
            .read(&returned_type)
            .read(&returned_data);
        EXPECT_EQ(returned_type, value_type_handle::messengers__status_online_away);
        EXPECT_NE(returned_data, &sent);
        EXPECT_NE(returned_data, nullptr);
        EXPECT_EQ(returned_data->m_device_type, device_type::personal_computer);
        EXPECT_EQ(returned_data->m_away_minutes, 1);
    }
}

TEST(DynamicStructures, InvalidReturnValue)
{
    auto *const library = library::get_instance();

    kernel kernel(library);
    auto domain1 = make_shared<local_domain_mock>(&kernel);

    const auto *const body = library->get_interface(interface_handle::messengers__contact);
    const method_base *get_status = nullptr;
    const method_id get_status_id = 0;
    ASSERT_EQ(body->get_supported_method(get_status_id, 0, get_status), method_lookup_result::implemented);

    auto host1 = domain1->register_module_side_host(body->get_handle(), 0);

    handle_type pack[100];
    object_handle releases[100];
    uint64_t release_counts[100];
    void* release_pointers[100];
    uint64_t release_count = 0;

    {
        handle_ptr<outbound_invocation<inbound_local_method_data>, module_to_kernel_invocation_handle> handle(
            domain1->get_outbound_invocations());
        auto* const request = static_cast<inbound_local_method_data*>(handle->get_inbound_method_data());
        request->set_request(host1, body->get_handle(), get_status_id, pack, releases, release_counts,
                             reinterpret_cast<char*>(release_pointers), &release_count);

        ASSERT_TRUE(domain1->start_invocation(handle.get_handle()));
        ASSERT_TRUE(handle);
        EXPECT_CALL(*domain1, send_request(_))
            .Times(1)
            .WillOnce(Return(domain_base::request_sending_status::pending_response));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());
        auto* local_data_passed = dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data());
        ASSERT_NE(local_data_passed, nullptr);

        object_handle target_object;
        interface_handle target_interface;
        method_id target_method;
        dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data())
            ->get_request_method_info(target_object, target_interface, target_method);
        ASSERT_EQ(target_object, host1);
        ASSERT_EQ(target_interface, body->get_handle());
        ASSERT_EQ(target_method, method_id(get_status_id));

        local_data_passed->access_return_value()
            .write(value_type_handle::messengers__device_type)
            .write(nullptr);
        local_data_passed->write_exception_data(exception_handle::invalid);

        auto callee_side_handle = handle->get_callee_side_handle();
        ASSERT_TRUE(domain1->process_response(callee_side_handle));
        EXPECT_CALL(*domain1,
                    pass_response_data_to_module(domain1->get_outbound_invocations()->access(handle.get_handle())))
            .Times(1)
            .WillOnce(Return(response_sending_status::wait_for_modules_confirmation));
        EXPECT_CALL(*domain1, call_returned_invalid_pack)
            .WillOnce([](invocation_request*, const exception_value& error) {
                EXPECT_NE(error.get_caller_side_handle(), exception_handle::invalid);
                EXPECT_EQ(error.get_kind(), exception_kind::system__invalid_value_type_parameter);
            });
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());

        exception_handle error_handle = static_cast<exception_handle>(static_cast<handle_type>(pack[0]));
        exception_kind error_kind = static_cast<exception_kind>(static_cast<handle_type>(pack[1]));
        ASSERT_NE(error_handle, exception_handle::invalid);
        ASSERT_EQ(error_kind, exception_kind::system__invalid_return_pack);
    }
}

TEST(DynamicStructures, InvalidParameter)
{
    auto *const library = library::get_instance();

    kernel kernel(library);
    auto domain1 = make_shared<local_domain_mock>(&kernel);

    const auto *const body = library->get_interface(interface_handle::messengers__status_handler);
    const method_base *truncate = nullptr;
    const method_id truncate_id = 0;
    ASSERT_EQ(body->get_supported_method(truncate_id, 0, truncate), method_lookup_result::implemented);

    auto host1 = domain1->register_module_side_host(body->get_handle(), 0);

    handle_type pack[100];
    object_handle releases[100];
    uint64_t release_counts[100];
    void* release_pointers[100];
    uint64_t release_count = 0;

    {
        pack_field_iterator(&pack[2], endian::native)
            .write(value_type_handle::messengers__device_type)
            .write(nullptr);

        handle_ptr<outbound_invocation<inbound_local_method_data>, module_to_kernel_invocation_handle> handle(
            domain1->get_outbound_invocations());
        auto* const request = static_cast<inbound_local_method_data*>(handle->get_inbound_method_data());
        request->set_request(host1, body->get_handle(), truncate_id, pack, releases, release_counts,
                             reinterpret_cast<char*>(release_pointers), &release_count);

        ASSERT_TRUE(domain1->start_invocation(handle.get_handle()));
        ASSERT_TRUE(handle);
        EXPECT_CALL(*domain1,
                    pass_response_data_to_module(domain1->get_outbound_invocations()->access(handle.get_handle())))
            .Times(1)
            .WillOnce(Return(response_sending_status::wait_for_modules_confirmation));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());

        exception_handle error_handle = static_cast<exception_handle>(static_cast<handle_type>(pack[0]));
        exception_kind error_kind = static_cast<exception_kind>(static_cast<handle_type>(pack[1]));
        ASSERT_NE(error_handle, exception_handle::invalid);
        ASSERT_EQ(error_kind, exception_kind::system__invalid_value_type_parameter);
    }
}

TEST(DynamicStructures, InvalidParameterChanged)
{
    auto *const library = library::get_instance();

    kernel kernel(library);
    auto domain1 = make_shared<local_domain_mock>(&kernel);

    const auto *const body = library->get_interface(interface_handle::messengers__status_handler);
    const method_base *truncate = nullptr;
    const method_id truncate_id = 0;
    ASSERT_EQ(body->get_supported_method(truncate_id, 0, truncate), method_lookup_result::implemented);

    auto host1 = domain1->register_module_side_host(body->get_handle(), 0);

    handle_type pack[100];
    object_handle releases[100];
    uint64_t release_counts[100];
    void* release_pointers[100];
    uint64_t release_count = 0;

    {
        pack_field_iterator(&pack[2], endian::native)
            .write(value_type_handle::messengers__status)
            .write(nullptr);

        handle_ptr<outbound_invocation<inbound_local_method_data>, module_to_kernel_invocation_handle> handle(
            domain1->get_outbound_invocations());
        auto* const request = static_cast<inbound_local_method_data*>(handle->get_inbound_method_data());
        request->set_request(host1, body->get_handle(), truncate_id, pack, releases, release_counts,
                             reinterpret_cast<char*>(release_pointers), &release_count);

        ASSERT_TRUE(domain1->start_invocation(handle.get_handle()));
        ASSERT_TRUE(handle);
        EXPECT_CALL(*domain1, send_request(_))
            .Times(1)
            .WillOnce(Return(domain_base::request_sending_status::pending_response));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());
        auto* local_data_passed = dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data());
        ASSERT_NE(local_data_passed, nullptr);

        object_handle target_object;
        interface_handle target_interface;
        method_id target_method;
        dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data())
            ->get_request_method_info(target_object, target_interface, target_method);
        ASSERT_EQ(target_object, host1);
        ASSERT_EQ(target_interface, body->get_handle());
        ASSERT_EQ(target_method, method_id(truncate_id));

        value_type_handle received_type = value_type_handle::invalid;
        void* received_data = nullptr;
        local_data_passed->access_parameters(0)
            .read(&received_type)
            .read(&received_data);
        EXPECT_EQ(received_type, value_type_handle::messengers__status);
        EXPECT_EQ(received_data, nullptr);

        local_data_passed->access_parameters(0)
            .write(value_type_handle::messengers__device_type)
            .write(nullptr);

        local_data_passed->write_exception_data(exception_handle::invalid);

        auto callee_side_handle = handle->get_callee_side_handle();
        ASSERT_TRUE(domain1->process_response(callee_side_handle));
        EXPECT_CALL(*domain1,
                    pass_response_data_to_module(domain1->get_outbound_invocations()->access(handle.get_handle())))
            .Times(1)
            .WillOnce(Return(response_sending_status::wait_for_modules_confirmation));
        EXPECT_CALL(*domain1, call_returned_invalid_pack)
            .WillOnce([](invocation_request*, const exception_value& error) {
                EXPECT_NE(error.get_caller_side_handle(), exception_handle::invalid);
                EXPECT_EQ(error.get_kind(), exception_kind::system__invalid_value_type_parameter);
            });
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());

        exception_handle error_handle = static_cast<exception_handle>(static_cast<handle_type>(pack[0]));
        exception_kind error_kind = static_cast<exception_kind>(static_cast<handle_type>(pack[1]));
        ASSERT_NE(error_handle, exception_handle::invalid);
        ASSERT_EQ(error_kind, exception_kind::system__invalid_return_pack);
    }
}
