#include <sivkernel/interface.h>
#include <sivkernel/method.h>
#include <sivkernel/kernel.h>
#include <sivkernel/generated/library.h>
#include <sivdk/exception_kind.h>
#include <sivkernel-mocks/domain_mock.h>
#include <sivdk/interface_handle.h>
#include <sivdk/value_type_handle.h>
#include <sivtypes/physics/mechanics/real_world_length_unit.h>
#include <gtest/gtest.h>
#include <gmock/gmock.h>
#include <string>
#include <unordered_set>

using namespace sivelkiria::kernel;
using namespace sivelkiria::kernel::generated;
using namespace sivtypes::physics::mechanics;
using namespace mocks;
using namespace sivelkiria::common;
using namespace data_pack;
using namespace std;
using namespace testing;


TEST(FixedStructures, ReturnValue)
{
    auto *const library = library::get_instance();

    kernel kernel(library);
    auto domain1 = make_shared<local_domain_mock>(&kernel);

    const auto *const body = library->get_interface(interface_handle::physics__mechanics__body);
    const method_base *get_width = nullptr;
    const method_id get_width_id = 0;
    ASSERT_EQ(body->get_supported_method(get_width_id, 0, get_width), method_lookup_result::implemented);

    auto host1 = domain1->register_module_side_host(body->get_handle(), 0);

    handle_type pack[100];
    object_handle releases[100];
    uint64_t release_counts[100];
    void* release_pointers[100];
    uint64_t release_count = 0;

    {
        handle_ptr<outbound_invocation<inbound_local_method_data>, module_to_kernel_invocation_handle> handle(
            domain1->get_outbound_invocations());
        auto* const request = static_cast<inbound_local_method_data*>(handle->get_inbound_method_data());
        request->set_request(host1, body->get_handle(), get_width_id, pack, releases, release_counts,
                             reinterpret_cast<char*>(release_pointers), &release_count);

        ASSERT_TRUE(domain1->start_invocation(handle.get_handle()));
        ASSERT_TRUE(handle);
        EXPECT_CALL(*domain1, send_request(_))
            .Times(1)
            .WillOnce(Return(domain_base::request_sending_status::pending_response));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());
        auto* local_data_passed = dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data());
        ASSERT_NE(local_data_passed, nullptr);

        object_handle target_object;
        interface_handle target_interface;
        method_id target_method;
        dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data())
            ->get_request_method_info(target_object, target_interface, target_method);
        ASSERT_EQ(target_object, host1);
        ASSERT_EQ(target_interface, body->get_handle());
        ASSERT_EQ(target_method, method_id(get_width_id));

        handle_type meters;
        pack_field_iterator(&meters, endian::native).write(real_world_length_unit::meter);
        local_data_passed->access_return_value()
            .write(static_cast<double>(2.58))
            .write(value_type_handle::physics__mechanics__real_world_length_unit_info)
            .write(&meters)
            .write(value_type_handle::physics__mechanics__rest_frame_length)
            .write(nullptr);

        local_data_passed->write_exception_data(exception_handle::invalid);

        auto callee_side_handle = handle->get_callee_side_handle();
        ASSERT_TRUE(domain1->process_response(callee_side_handle));
        EXPECT_CALL(*domain1,
                    pass_response_data_to_module(domain1->get_outbound_invocations()->access(handle.get_handle())))
            .Times(1)
            .WillOnce(Return(response_sending_status::wait_for_modules_confirmation));
        EXPECT_CALL(*domain1, request_module_free_response_dynamic_data(callee_side_handle, _)).Times(1);
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());

        exception_handle error_handle = static_cast<exception_handle>(static_cast<handle_type>(pack[0]));
        exception_kind error_kind = static_cast<exception_kind>(static_cast<handle_type>(pack[1]));
        ASSERT_EQ(error_handle, exception_handle::invalid);
        ASSERT_EQ(error_kind, static_cast<exception_kind>(static_cast<uint64_t>(invalid_interface_handle)));

        double length = 0;
        value_type_handle length_unit_type = value_type_handle::invalid;
        handle_type* length_unit_info = nullptr;
        value_type_handle unit_conversion_context_type = value_type_handle::invalid;
        handle_type* unit_conversion_context_info = nullptr;

        pack_field_const_iterator(&pack[2], endian::native)
            .read(&length)
            .read(&length_unit_type)
            .read(&length_unit_info)
            .read(&unit_conversion_context_type)
            .read(&unit_conversion_context_info);

        EXPECT_EQ(length, 2.58);
        EXPECT_EQ(length_unit_type, value_type_handle::physics__mechanics__real_world_length_unit_info);
        EXPECT_NE(length_unit_info, nullptr);
        EXPECT_NE(length_unit_info, &meters);
        EXPECT_EQ(unit_conversion_context_type, value_type_handle::physics__mechanics__rest_frame_length);
        EXPECT_EQ(unit_conversion_context_info, nullptr);

        real_world_length_unit unit = real_world_length_unit::astronomical_unit;
        pack_field_const_iterator(length_unit_info, endian::native).read(&unit);
        EXPECT_EQ(unit, real_world_length_unit::meter);
    }

    {
        handle_ptr<outbound_invocation<inbound_local_method_data>, module_to_kernel_invocation_handle> handle(
            domain1->get_outbound_invocations());
        auto* const request = static_cast<inbound_local_method_data*>(handle->get_inbound_method_data());
        request->set_request(host1, body->get_handle(), get_width_id, pack, releases, release_counts,
                             reinterpret_cast<char*>(release_pointers), &release_count);

        ASSERT_TRUE(domain1->start_invocation(handle.get_handle()));
        ASSERT_TRUE(handle);
        EXPECT_CALL(*domain1, send_request(_))
            .Times(1)
            .WillOnce(Return(domain_base::request_sending_status::pending_response));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());
        auto* local_data_passed = dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data());
        ASSERT_NE(local_data_passed, nullptr);

        object_handle target_object;
        interface_handle target_interface;
        method_id target_method;
        dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data())
            ->get_request_method_info(target_object, target_interface, target_method);
        ASSERT_EQ(target_object, host1);
        ASSERT_EQ(target_interface, body->get_handle());
        ASSERT_EQ(target_method, method_id(get_width_id));

        handle_type meters;
        pack_field_iterator(&meters, endian::native).write(real_world_length_unit::meter);
        local_data_passed->access_return_value()
            .write(static_cast<double>(2.58))
            .write(value_type_handle::invalid)
            .write(nullptr)
            .write(value_type_handle::invalid)
            .write(nullptr);

        local_data_passed->write_exception_data(exception_handle::invalid);

        auto callee_side_handle = handle->get_callee_side_handle();
        ASSERT_TRUE(domain1->process_response(callee_side_handle));
        EXPECT_CALL(*domain1,
                    pass_response_data_to_module(domain1->get_outbound_invocations()->access(handle.get_handle())))
            .Times(1)
            .WillOnce(Return(response_sending_status::wait_for_modules_confirmation));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());

        exception_handle error_handle = static_cast<exception_handle>(static_cast<handle_type>(pack[0]));
        exception_kind error_kind = static_cast<exception_kind>(static_cast<handle_type>(pack[1]));
        ASSERT_EQ(error_handle, exception_handle::invalid);
        ASSERT_EQ(error_kind, static_cast<exception_kind>(static_cast<uint64_t>(invalid_interface_handle)));

        double length = 0;
        value_type_handle length_unit_type = value_type_handle::invalid;
        handle_type* length_unit_info = nullptr;
        value_type_handle unit_conversion_context_type = value_type_handle::invalid;
        handle_type* unit_conversion_context_info = nullptr;

        pack_field_const_iterator(&pack[2], endian::native)
            .read(&length)
            .read(&length_unit_type)
            .read(&length_unit_info)
            .read(&unit_conversion_context_type)
            .read(&unit_conversion_context_info);

        EXPECT_EQ(length, 2.58);
        EXPECT_EQ(length_unit_type, value_type_handle::invalid);
        EXPECT_EQ(length_unit_info, nullptr);
        EXPECT_EQ(unit_conversion_context_type, value_type_handle::invalid);
        EXPECT_EQ(unit_conversion_context_info, nullptr);
    }
}

TEST(FixedStructures, InputParameter)
{
    auto *const library = library::get_instance();

    kernel kernel(library);
    auto domain1 = make_shared<local_domain_mock>(&kernel);

    const auto *const body = library->get_interface(interface_handle::graphics__graphics_2d__canvas);
    const method_base *fit_size = nullptr;
    const method_id fit_size_id = 0;
    ASSERT_EQ(body->get_supported_method(fit_size_id, 0, fit_size), method_lookup_result::implemented);

    auto host1 = domain1->register_module_side_host(body->get_handle(), 0);

    handle_type pack[100];
    object_handle releases[100];
    uint64_t release_counts[100];
    void* release_pointers[100];
    uint64_t release_count = 0;

    {
        pack_field_iterator(&pack[2], endian::native)
            .write(uint64_t(123))
            .write(uint64_t(456));

        handle_ptr<outbound_invocation<inbound_local_method_data>, module_to_kernel_invocation_handle> handle(
            domain1->get_outbound_invocations());
        auto* const request = static_cast<inbound_local_method_data*>(handle->get_inbound_method_data());
        request->set_request(host1, body->get_handle(), fit_size_id, pack, releases, release_counts,
                             reinterpret_cast<char*>(release_pointers), &release_count);

        ASSERT_TRUE(domain1->start_invocation(handle.get_handle()));
        ASSERT_TRUE(handle);
        EXPECT_CALL(*domain1, send_request(_))
            .Times(1)
            .WillOnce(Return(domain_base::request_sending_status::pending_response));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());
        auto* local_data_passed = dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data());
        ASSERT_NE(local_data_passed, nullptr);

        object_handle target_object;
        interface_handle target_interface;
        method_id target_method;
        dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data())
            ->get_request_method_info(target_object, target_interface, target_method);
        ASSERT_EQ(target_object, host1);
        ASSERT_EQ(target_interface, body->get_handle());
        ASSERT_EQ(target_method, method_id(fit_size_id));

        uint64_t width = 0;
        uint64_t height = 0;
        local_data_passed->access_parameters(0)
            .read(&width)
            .read(&height);
        EXPECT_EQ(width, 123);
        EXPECT_EQ(height, 456);

        local_data_passed->write_exception_data(exception_handle::invalid);

        auto callee_side_handle = handle->get_callee_side_handle();
        ASSERT_TRUE(domain1->process_response(callee_side_handle));
        EXPECT_CALL(*domain1,
                    pass_response_data_to_module(domain1->get_outbound_invocations()->access(handle.get_handle())))
            .Times(1)
            .WillOnce(Return(response_sending_status::wait_for_modules_confirmation));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());

        exception_handle error_handle = static_cast<exception_handle>(static_cast<handle_type>(pack[0]));
        exception_kind error_kind = static_cast<exception_kind>(static_cast<handle_type>(pack[1]));
        ASSERT_EQ(error_handle, exception_handle::invalid);
        ASSERT_EQ(error_kind, static_cast<exception_kind>(static_cast<uint64_t>(invalid_interface_handle)));

        EXPECT_EQ(pack[2], 123);
        EXPECT_EQ(pack[3], 456);
    }
}

TEST(FixedStructures, InputParameterChanged)
{
    auto *const library = library::get_instance();

    kernel kernel(library);
    auto domain1 = make_shared<local_domain_mock>(&kernel);

    const auto *const body = library->get_interface(interface_handle::graphics__graphics_2d__canvas);
    const method_base *fit_size = nullptr;
    const method_id fit_size_id = 0;
    ASSERT_EQ(body->get_supported_method(fit_size_id, 0, fit_size), method_lookup_result::implemented);

    auto host1 = domain1->register_module_side_host(body->get_handle(), 0);

    handle_type pack[100];
    object_handle releases[100];
    uint64_t release_counts[100];
    void* release_pointers[100];
    uint64_t release_count = 0;

    {
        pack_field_iterator(&pack[2], endian::native)
            .write(uint64_t(123))
            .write(uint64_t(456));

        handle_ptr<outbound_invocation<inbound_local_method_data>, module_to_kernel_invocation_handle> handle(
            domain1->get_outbound_invocations());
        auto* const request = static_cast<inbound_local_method_data*>(handle->get_inbound_method_data());
        request->set_request(host1, body->get_handle(), fit_size_id, pack, releases, release_counts,
                             reinterpret_cast<char*>(release_pointers), &release_count);

        ASSERT_TRUE(domain1->start_invocation(handle.get_handle()));
        ASSERT_TRUE(handle);
        EXPECT_CALL(*domain1, send_request(_))
            .Times(1)
            .WillOnce(Return(domain_base::request_sending_status::pending_response));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());
        auto* local_data_passed = dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data());
        ASSERT_NE(local_data_passed, nullptr);

        object_handle target_object;
        interface_handle target_interface;
        method_id target_method;
        dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data())
            ->get_request_method_info(target_object, target_interface, target_method);
        ASSERT_EQ(target_object, host1);
        ASSERT_EQ(target_interface, body->get_handle());
        ASSERT_EQ(target_method, method_id(fit_size_id));

        uint64_t width = 0;
        uint64_t height = 0;
        local_data_passed->access_parameters(0)
            .read(&width)
            .read(&height);
        EXPECT_EQ(width, 123);
        EXPECT_EQ(height, 456);

        local_data_passed->access_parameters(0)
            .write(uint64_t(789))
            .write(uint64_t(1000));

        local_data_passed->write_exception_data(exception_handle::invalid);

        auto callee_side_handle = handle->get_callee_side_handle();
        ASSERT_TRUE(domain1->process_response(callee_side_handle));
        EXPECT_CALL(*domain1,
                    pass_response_data_to_module(domain1->get_outbound_invocations()->access(handle.get_handle())))
            .Times(1)
            .WillOnce(Return(response_sending_status::wait_for_modules_confirmation));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());

        exception_handle error_handle = static_cast<exception_handle>(static_cast<handle_type>(pack[0]));
        exception_kind error_kind = static_cast<exception_kind>(static_cast<handle_type>(pack[1]));
        ASSERT_EQ(error_handle, exception_handle::invalid);
        ASSERT_EQ(error_kind, static_cast<exception_kind>(static_cast<uint64_t>(invalid_interface_handle)));

        EXPECT_EQ(pack[2], 789);
        EXPECT_EQ(pack[3], 1000);
    }
}

