#define _SILENCE_CXX17_CODECVT_HEADER_DEPRECATION_WARNING 1

#include "console.h"
#include <sivtypes/text/console.h>
#include <sivtypes/text/writer.h>
#include <iostream>
#include <codecvt>
#include <string>
#include <locale>
#include <mutex>

using namespace std;
using namespace sivdk;
using namespace sivtypes::text;

namespace {

mutex console_lock;

}

console_implementation::console_implementation()
    : object(interface_handle::text__console, get_interface_version())
{}

shared_ptr<::sivtypes::text::writer> console_implementation::get_writer()
{
    class writer : public ::sivtypes::text::writer
    {
    public:
        writer()
            : object(interface_handle::text__writer, get_interface_version())
        {
    #ifdef WIN32
            wcout.imbue(locale("rus_rus.866"));
    #else
            setlocale(LC_ALL, "");
    #endif
        }
        virtual ~writer() = default;

        void write(const u16string& text) override
        {
            wstring_convert<codecvt_utf16<wchar_t, 0x10ffff, std::little_endian>, wchar_t> conv;
            const auto converted = conv.from_bytes(reinterpret_cast<const char*>(&text[0]),
                                                   reinterpret_cast<const char*>(&text[0] + text.size()));
            std::unique_lock<mutex> lock(console_lock);
            wcout << converted;
        }
    };

    return make_shared<writer>();
}

uint64_t console_implementation::get_extra_thread_count()
{
    return 0;
}

void console_implementation::run_extra_thread(uint64_t)
{}

void console_implementation::stop_extra_threads()
{}
