#include "filter.h"
#include <sivtypes/messengers/receiver.h>
#include <sivtypes/messengers/message.h>
#include <sivdk/interface_handle.h>

using namespace std;
using namespace sivdk;
using namespace sivtypes::messengers;

filter_implementation::filter_implementation()
    : object(interface_handle::messengers__filter, get_interface_version())
{}

shared_ptr<receiver> filter_implementation::start_filtering(const shared_ptr<receiver>& output)
{
    class message_implementation : public message
    {
    public:
        message_implementation(const u16string& sender_name, const u16string& text)
            : object(interface_handle::messengers__message, get_interface_version())
            , m_sender_name(sender_name)
            , m_text(text)
        {}
        virtual ~message_implementation() = default;

        u16string get_text() override
        {
            return m_text;
        }
        u16string get_sender_name() override
        {
            return m_sender_name;
        }

    private:
        const u16string m_sender_name;
        const u16string m_text;
    };

    class proxy : public receiver
    {
    public:
        proxy(const shared_ptr<receiver>& output)
            : object(interface_handle::messengers__receiver, get_interface_version())
            , m_output(output)
        {}
        virtual ~proxy() = default;

        void receive(const shared_ptr<message>& received) override
        {
            received->get_text__async([=, this](auto getter) {
                auto text = getter.get();
                auto bad_word_at = text.find(u"hate");
                if (bad_word_at == u16string::npos)
                {
                    m_output->receive(received);
                    return;
                }

                do
                {
                    text[bad_word_at] = u'l';
                    text[bad_word_at + 1] = u'o';
                    text[bad_word_at + 2] = u'v';
                    bad_word_at = text.find(u"hate", bad_word_at + 4);
                } while (bad_word_at != u16string::npos);

                m_output->receive(make_shared<message_implementation>(received->get_sender_name(), text));
            });
        }

    private:
        const shared_ptr<receiver> m_output;
    };

    return make_shared<proxy>(output);
}

uint64_t filter_implementation::get_extra_thread_count()
{
    return 0;
}

void filter_implementation::run_extra_thread(uint64_t)
{}

void filter_implementation::stop_extra_threads()
{}
