#include "censorer.h"
#include <sivtypes/text/censorer.h>
#include <sivtypes/text/writer.h>

using namespace std;
using namespace sivdk;
using namespace sivtypes::text;

censorer_implementation::censorer_implementation()
    : object(interface_handle::text__censorer, get_interface_version())
{}

shared_ptr<writer> censorer_implementation::get_censorer(const shared_ptr<writer>& output)
{
    static const u16string bad_word = u"Mr. Bossman";
    static const u16string good_word = u"Mr. *******";

    class processor : public writer
    {
    public:
        processor(const shared_ptr<writer>& output)
            : object(interface_handle::text__writer, get_interface_version())
            , m_output(output)
        {}
        virtual ~processor() = default;

        void write(const u16string& text) override
        {
            u16string::size_type last = 0;
            u16string result;

            while (true)
            {
                const auto bad_word_at = text.find(bad_word, last);
                if (bad_word_at == u16string::npos)
                    break;

                result += text.substr(last, bad_word_at - last);
                result += good_word;
                last = bad_word_at + bad_word.size();
            }

            result += text.substr(last);
            m_output->write(result);
        }

    private:
        const shared_ptr<writer> m_output;
    };

    return make_shared<processor>(output);
}

uint64_t censorer_implementation::get_extra_thread_count()
{
    return 0;
}

void censorer_implementation::run_extra_thread(uint64_t)
{}

void censorer_implementation::stop_extra_threads()
{}
