#include "proxy_provider.h"
#include <sivtypes/messengers/message_source_proxy.h>
#include <sivdk/interface_handle.h>
#include <vector>
#include <mutex>

using namespace std;
using namespace sivdk;
using namespace sivtypes::messengers;

proxy_provider::proxy_provider()
    : object(interface_handle::messengers__message_source_proxy_provider, get_interface_version())
{}

shared_ptr<message_source_proxy> proxy_provider::create_proxy()
{
    class proxy : public message_source_proxy
    {
    public:
        proxy()
            : object(interface_handle::messengers__message_source_proxy, get_interface_version())
        {}
        virtual ~proxy() = default;

        void push(const shared_ptr<message>& received) override
        {
            emit__new_message(received);

            unique_lock<mutex> lock(m_messages_lock);
            m_messages.push_back(received);
        }
        uint64_t get_message_count() override
        {
            unique_lock<mutex> lock(m_messages_lock);
            return m_messages.size();
        }
        shared_ptr<message> get_message_by_index(uint64_t index) override
        {
            unique_lock<mutex> lock(m_messages_lock);
            if (index < m_messages.size())
                return m_messages[static_cast<size_type>(index)];
            return nullptr;
        }

    private:
        vector<shared_ptr<message>> m_messages;
        mutex m_messages_lock;
    };

    return make_shared<proxy>();
}

uint64_t proxy_provider::get_extra_thread_count()
{
    return 0;
}

void proxy_provider::run_extra_thread(uint64_t)
{}

void proxy_provider::stop_extra_threads()
{}
