#include <sivdk/outgoing_call.h>
#include <sivdk/module_functions.h>
#include <sivdk/call_data/proxy.h>
#include <gtest/gtest.h>
#include <sivdk-mocks/kernel_mock.h>
#include <sivdk-mocks/method_defines.h>

#include <sivtypes/messengers/contact.h>
#include <sivtypes/messengers/device_type.h>
#include <sivtypes/messengers/status.h>
#include <sivtypes/messengers/status_online.h>
#include <sivtypes/messengers/status_offline.h>
#include <sivtypes/messengers/status_online_active.h>
#include <sivtypes/messengers/status_online_away.h>
#include <sivtypes/messengers/status_handler.h>
#include <sivmocks/messengers/contact.h>
#include <sivmocks/messengers/status_handler.h>
#include <sivdk/exception_kind.h>
#include <sivdk/value_type_handle.h>

using namespace sivdk;
using namespace call_data;
using namespace std;
using namespace mocks;
using namespace testing;
using namespace sivtypes::messengers;


TEST(DynamicStructure, OutboundCallReturnValue)
{
    object_proxy_handle_holder proxy_handle(object_handle(1));
    auto a = std::make_shared<contact__proxy>(&proxy_handle);

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, invoke_synchronously(object_handle(1), interface_handle::messengers__contact, GET_METHOD_ID(contact::get_status), _, _, _, _, _, _))
            .Times(1)
            .WillOnce([](object_handle, interface_handle, method_id,
                         void* data, object_handle* released, uint64_t* reference_counts,
                         void** released_pointers, uint64_t* released_count,
                         module_to_kernel_invocation_handle* handle)
                {
                    EXPECT_NE(data, nullptr);
                    EXPECT_EQ(released, nullptr);
                    EXPECT_EQ(reference_counts, nullptr);
                    EXPECT_EQ(released_pointers, nullptr);
                    EXPECT_NE(released_count, nullptr);
                    EXPECT_EQ(*released_count, 0);
                    EXPECT_NE(handle, nullptr);

                    *handle = module_to_kernel_invocation_handle(123);
                    pack_field_iterator(data)
                        .write(exception_handle::invalid)
                        .write(exception_kind::invalid)
                        .write(value_type_handle::messengers__status)
                        .write(nullptr);

                    return invocation_status::success;
                }
            );
        auto status = a->get_status();

        EXPECT_NE(status, nullptr);
        EXPECT_EQ(status->get_type(), value_type_handle::messengers__status);
    }

    {
        struct {
            alignas(handle_type) uint64_t m_minutes;
        } offline_body {
            50
        };

        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, invoke_synchronously(object_handle(1), interface_handle::messengers__contact, GET_METHOD_ID(contact::get_status), _, _, _, _, _, _))
            .Times(1)
            .WillOnce([&offline_body](object_handle, interface_handle, method_id,
                         void* data, object_handle* released, uint64_t* reference_counts,
                         void** released_pointers, uint64_t* released_count,
                         module_to_kernel_invocation_handle* handle)
                {
                    EXPECT_NE(data, nullptr);
                    EXPECT_EQ(released, nullptr);
                    EXPECT_EQ(reference_counts, nullptr);
                    EXPECT_EQ(released_pointers, nullptr);
                    EXPECT_NE(released_count, nullptr);
                    EXPECT_EQ(*released_count, 0);
                    EXPECT_NE(handle, nullptr);

                    *handle = module_to_kernel_invocation_handle(123);
                    pack_field_iterator(data)
                        .write(exception_handle::invalid)
                        .write(exception_kind::invalid)
                        .write(value_type_handle::messengers__status_offline)
                        .write(&offline_body);

                    return invocation_status::success;
                }
            );
        EXPECT_CALL(kernel, release_kernel_side_containers(module_to_kernel_invocation_handle(123)));
        auto status = dynamic_pointer_cast<status_offline>(a->get_status());

        EXPECT_NE(status, nullptr);
        EXPECT_EQ(status->get_type(), value_type_handle::messengers__status_offline);
        EXPECT_EQ(status->minutes, 50);
    }

    {
        struct {
            alignas(handle_type) device_type m_device_type;
            alignas(handle_type) uint64_t m_minutes;
        } away_body {
            device_type::mobile_device,
            15
        };

        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, invoke_synchronously(object_handle(1), interface_handle::messengers__contact, GET_METHOD_ID(contact::get_status), _, _, _, _, _, _))
            .Times(1)
            .WillOnce([&away_body](object_handle, interface_handle, method_id,
                         void* data, object_handle* released, uint64_t* reference_counts,
                         void** released_pointers, uint64_t* released_count,
                         module_to_kernel_invocation_handle* handle)
                {
                    EXPECT_NE(data, nullptr);
                    EXPECT_EQ(released, nullptr);
                    EXPECT_EQ(reference_counts, nullptr);
                    EXPECT_EQ(released_pointers, nullptr);
                    EXPECT_NE(released_count, nullptr);
                    EXPECT_EQ(*released_count, 0);
                    EXPECT_NE(handle, nullptr);

                    *handle = module_to_kernel_invocation_handle(123);
                    pack_field_iterator(data)
                        .write(exception_handle::invalid)
                        .write(exception_kind::invalid)
                        .write(value_type_handle::messengers__status_online_away)
                        .write(&away_body);

                    return invocation_status::success;
                }
            );
        EXPECT_CALL(kernel, release_kernel_side_containers(module_to_kernel_invocation_handle(123)));
        auto status = dynamic_pointer_cast<status_online_away>(a->get_status());

        EXPECT_NE(status, nullptr);
        EXPECT_EQ(status->get_type(), value_type_handle::messengers__status_online_away);
        EXPECT_EQ(status->device_type, device_type::mobile_device);
        EXPECT_EQ(status->minutes, 15);
    }

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, invoke_synchronously(object_handle(1), interface_handle::messengers__contact, GET_METHOD_ID(contact::get_status), _, _, _, _, _, _))
            .Times(1)
            .WillOnce([](object_handle, interface_handle, method_id,
                         void* data, object_handle* released, uint64_t* reference_counts,
                         void** released_pointers, uint64_t* released_count,
                         module_to_kernel_invocation_handle* handle)
                {
                    EXPECT_NE(data, nullptr);
                    EXPECT_EQ(released, nullptr);
                    EXPECT_EQ(reference_counts, nullptr);
                    EXPECT_EQ(released_pointers, nullptr);
                    EXPECT_NE(released_count, nullptr);
                    EXPECT_EQ(*released_count, 0);
                    EXPECT_NE(handle, nullptr);

                    *handle = module_to_kernel_invocation_handle(123);
                    pack_field_iterator(data)
                        .write(exception_handle::invalid)
                        .write(exception_kind::invalid)
                        .write(value_type_handle::invalid)
                        .write(nullptr);

                    return invocation_status::success;
                }
            );
        auto status = a->get_status();
        EXPECT_EQ(status, nullptr);
    }

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, release_object_handle(object_handle(1))).Times(1).WillOnce(Return(true));
        a = nullptr;
    }
}

TEST(DynamicStructure, InboundCallReturnValue)
{
    std::shared_ptr<contact__mock> a;

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, register_host_object(interface_handle::messengers__contact, 0, _))
            .WillOnce(Return(object_handle(1234)));
        a = std::make_shared<contact__mock>();
    }

    EXPECT_EQ(a->get_handle(), object_handle(1234));

    {
        struct
        {
            alignas(handle_type) exception_throw_mode m_mode;
            alignas(handle_type) handle_type m_handle;
            alignas(handle_type) value_type_handle m_status_type;
            alignas(handle_type) void* m_status_data;
        } data {
            exception_throw_mode::no_exception,
            invalid_handle,
            value_type_handle::invalid,
            nullptr
        };
        void* invocation_pointer = nullptr;

        shared_ptr<status> result = nullptr;
        EXPECT_CALL(*a, get_status()).WillOnce(Return(result));
        handle_call(object_handle(1234), static_cast<object*>(a.get()), interface_handle::messengers__contact,
                    GET_METHOD_ID(contact::get_status), &data, kernel_to_module_invocation_handle(1),
                    &invocation_pointer);

        EXPECT_EQ(data.m_mode, exception_throw_mode::no_exception);
        EXPECT_EQ(data.m_handle, invalid_handle);

        EXPECT_EQ(data.m_status_type, value_type_handle::invalid);
        EXPECT_EQ(data.m_status_data, nullptr);

        EXPECT_EQ(invocation_pointer, nullptr);
    }

    {
        struct
        {
            alignas(handle_type) exception_throw_mode m_mode;
            alignas(handle_type) handle_type m_handle;
            alignas(handle_type) value_type_handle m_status_type;
            alignas(handle_type) void* m_status_data;
        } data {
            exception_throw_mode::no_exception,
            invalid_handle,
            value_type_handle::invalid,
            nullptr
        };
        void* invocation_pointer = nullptr;

        auto result = make_shared<status>();
        EXPECT_CALL(*a, get_status()).WillOnce(Return(result));
        handle_call(object_handle(1234), static_cast<object*>(a.get()), interface_handle::messengers__contact,
                    GET_METHOD_ID(contact::get_status), &data, kernel_to_module_invocation_handle(1),
                    &invocation_pointer);

        EXPECT_EQ(data.m_mode, exception_throw_mode::no_exception);
        EXPECT_EQ(data.m_handle, invalid_handle);

        EXPECT_EQ(data.m_status_type, value_type_handle::messengers__status);
        EXPECT_EQ(data.m_status_data, nullptr);

        EXPECT_EQ(invocation_pointer, nullptr);
    }

    {
        struct away_data_type
        {
            alignas(handle_type) device_type m_device_type;
            alignas(handle_type) uint64_t m_away_minutes;
        };

        struct
        {
            alignas(handle_type) exception_throw_mode m_mode;
            alignas(handle_type) handle_type m_handle;
            alignas(handle_type) value_type_handle m_status_type;
            alignas(handle_type) void* m_status_data;
        } data {
            exception_throw_mode::no_exception,
            invalid_handle,
            value_type_handle::invalid,
            nullptr
        };
        void* invocation_pointer = nullptr;

        auto result = make_shared<status_online_away>(device_type::personal_computer, 180);
        EXPECT_CALL(*a, get_status()).WillOnce(Return(result));
        handle_call(object_handle(1234), static_cast<object*>(a.get()), interface_handle::messengers__contact,
                    GET_METHOD_ID(contact::get_status), &data, kernel_to_module_invocation_handle(1),
                    &invocation_pointer);

        EXPECT_EQ(data.m_mode, exception_throw_mode::no_exception);
        EXPECT_EQ(data.m_handle, invalid_handle);

        EXPECT_EQ(data.m_status_type, value_type_handle::messengers__status_online_away);
        EXPECT_NE(data.m_status_data, nullptr);

        auto away_data = reinterpret_cast<away_data_type*>(data.m_status_data);
        EXPECT_EQ(away_data->m_device_type, device_type::personal_computer);
        EXPECT_EQ(away_data->m_away_minutes, 180);

        EXPECT_NE(invocation_pointer, nullptr);
        release_inbound_call_data(kernel_to_module_invocation_handle(1), invocation_pointer);
    }

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, release_object_handle(object_handle(1234))).Times(1).WillOnce(Return(true));
        a = nullptr;
    }
}

TEST(DynamicStructure, OutboundCallInputParameter)
{
    object_proxy_handle_holder proxy_handle(object_handle(1));
    auto a = std::make_shared<status_handler__proxy>(&proxy_handle);

    struct fields_type
    {
        alignas(handle_type) exception_throw_mode m_mode;
        alignas(handle_type) handle_type m_handle;
        alignas(handle_type) value_type_handle m_status_type;
        alignas(handle_type) void* m_status_data;
    };

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, invoke_synchronously(object_handle(1), interface_handle::messengers__status_handler, GET_METHOD_ID(status_handler::truncate), _, _, _, _, _, _))
            .Times(1)
            .WillOnce([](object_handle, interface_handle, method_id,
                         void* data, object_handle* released, uint64_t* reference_counts,
                         void** released_pointers, uint64_t* released_count,
                         module_to_kernel_invocation_handle* handle)
                {
                    EXPECT_NE(data, nullptr);
                    EXPECT_EQ(released, nullptr);
                    EXPECT_EQ(reference_counts, nullptr);
                    EXPECT_EQ(released_pointers, nullptr);
                    EXPECT_NE(released_count, nullptr);
                    EXPECT_EQ(*released_count, 0);
                    EXPECT_NE(handle, nullptr);

                    auto *const fields = reinterpret_cast<fields_type*>(data);

                    EXPECT_EQ(fields->m_status_type, value_type_handle::invalid);
                    EXPECT_EQ(fields->m_status_data, nullptr);

                    *handle = module_to_kernel_invocation_handle(123);
                    pack_field_iterator(data)
                        .write(exception_handle::invalid)
                        .write(exception_kind::invalid);

                    return invocation_status::success;
                }
            );

        shared_ptr<status> argument = nullptr;
        a->truncate(argument);

        EXPECT_EQ(argument, nullptr);
    }

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, invoke_synchronously(object_handle(1), interface_handle::messengers__status_handler, GET_METHOD_ID(status_handler::truncate), _, _, _, _, _, _))
            .Times(1)
            .WillOnce([](object_handle, interface_handle, method_id,
                         void* data, object_handle* released, uint64_t* reference_counts,
                         void** released_pointers, uint64_t* released_count,
                         module_to_kernel_invocation_handle* handle)
                {
                    EXPECT_NE(data, nullptr);
                    EXPECT_EQ(released, nullptr);
                    EXPECT_EQ(reference_counts, nullptr);
                    EXPECT_EQ(released_pointers, nullptr);
                    EXPECT_NE(released_count, nullptr);
                    EXPECT_EQ(*released_count, 0);
                    EXPECT_NE(handle, nullptr);

                    auto *const fields = reinterpret_cast<fields_type*>(data);

                    EXPECT_EQ(fields->m_status_type, value_type_handle::messengers__status);
                    EXPECT_EQ(fields->m_status_data, nullptr);

                    *handle = module_to_kernel_invocation_handle(123);
                    pack_field_iterator(data)
                        .write(exception_handle::invalid)
                        .write(exception_kind::invalid);

                    return invocation_status::success;
                }
            );

        shared_ptr<status> argument = make_shared<status>();
        a->truncate(argument);

        EXPECT_NE(argument, nullptr);
        EXPECT_EQ(argument->get_type(), value_type_handle::messengers__status);
    }

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, invoke_synchronously(object_handle(1), interface_handle::messengers__status_handler, GET_METHOD_ID(status_handler::truncate), _, _, _, _, _, _))
            .Times(1)
            .WillOnce([](object_handle, interface_handle, method_id,
                         void* data, object_handle* released, uint64_t* reference_counts,
                         void** released_pointers, uint64_t* released_count,
                         module_to_kernel_invocation_handle* handle)
                {
                    EXPECT_NE(data, nullptr);
                    EXPECT_EQ(released, nullptr);
                    EXPECT_EQ(reference_counts, nullptr);
                    EXPECT_EQ(released_pointers, nullptr);
                    EXPECT_NE(released_count, nullptr);
                    EXPECT_EQ(*released_count, 0);
                    EXPECT_NE(handle, nullptr);

                    auto *const fields = reinterpret_cast<fields_type*>(data);

                    EXPECT_EQ(fields->m_status_type, value_type_handle::messengers__status_offline);
                    EXPECT_NE(fields->m_status_data, nullptr);

                    uint64_t *const offline_minutes = reinterpret_cast<uint64_t*>(fields->m_status_data);
                    EXPECT_EQ(*offline_minutes, 240);

                    *handle = module_to_kernel_invocation_handle(123);
                    pack_field_iterator(data)
                        .write(exception_handle::invalid)
                        .write(exception_kind::invalid);

                    return invocation_status::success;
                }
            );

        shared_ptr<status> argument = make_shared<status_offline>(240);
        a->truncate(argument);

        EXPECT_NE(argument, nullptr);
        EXPECT_EQ(argument->get_type(), value_type_handle::messengers__status_offline);
        EXPECT_EQ(dynamic_pointer_cast<status_offline>(argument)->minutes, 240);
    }

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, invoke_synchronously(object_handle(1), interface_handle::messengers__status_handler, GET_METHOD_ID(status_handler::truncate), _, _, _, _, _, _))
            .Times(1)
            .WillOnce([](object_handle, interface_handle, method_id,
                         void* data, object_handle* released, uint64_t* reference_counts,
                         void** released_pointers, uint64_t* released_count,
                         module_to_kernel_invocation_handle* handle)
                {
                    EXPECT_NE(data, nullptr);
                    EXPECT_EQ(released, nullptr);
                    EXPECT_EQ(reference_counts, nullptr);
                    EXPECT_EQ(released_pointers, nullptr);
                    EXPECT_NE(released_count, nullptr);
                    EXPECT_EQ(*released_count, 0);
                    EXPECT_NE(handle, nullptr);

                    auto *const fields = reinterpret_cast<fields_type*>(data);

                    EXPECT_EQ(fields->m_status_type, value_type_handle::messengers__status_online_away);
                    EXPECT_NE(fields->m_status_data, nullptr);

                    struct away_fields_type {
                        alignas(uint64_t) device_type m_device_type;
                        alignas(uint64_t) uint64_t m_minutes_away;
                    } *const away_fields = reinterpret_cast<away_fields_type*>(fields->m_status_data);

                    EXPECT_EQ(away_fields->m_device_type, device_type::bot);
                    EXPECT_EQ(away_fields->m_minutes_away, 0);

                    *handle = module_to_kernel_invocation_handle(123);
                    pack_field_iterator(data)
                        .write(exception_handle::invalid)
                        .write(exception_kind::invalid);

                    return invocation_status::success;
                }
            );

        shared_ptr<status> argument = make_shared<status_online_away>(device_type::bot, 0);
        a->truncate(argument);

        EXPECT_NE(argument, nullptr);
        EXPECT_EQ(argument->get_type(), value_type_handle::messengers__status_online_away);
        EXPECT_EQ(dynamic_pointer_cast<status_online_away>(argument)->device_type, device_type::bot);
        EXPECT_EQ(dynamic_pointer_cast<status_online_away>(argument)->minutes, 0);
    }

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, release_object_handle(object_handle(1))).Times(1).WillOnce(Return(true));
        a = nullptr;
    }
}

TEST(DynamicStructure, OutboundCallParameterChange)
{
    object_proxy_handle_holder proxy_handle(object_handle(1));
    auto a = std::make_shared<status_handler__proxy>(&proxy_handle);

    struct fields_type
    {
        alignas(handle_type) exception_throw_mode m_mode;
        alignas(handle_type) handle_type m_handle;
        alignas(handle_type) value_type_handle m_status_type;
        alignas(handle_type) void* m_status_data;
    };
    struct away_fields_type {
        alignas(uint64_t) device_type m_device_type;
        alignas(uint64_t) uint64_t m_minutes_away;
    };

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, invoke_synchronously(object_handle(1), interface_handle::messengers__status_handler, GET_METHOD_ID(status_handler::truncate), _, _, _, _, _, _))
            .Times(1)
            .WillOnce([](object_handle, interface_handle, method_id,
                         void* data, object_handle* released, uint64_t* reference_counts,
                         void** released_pointers, uint64_t* released_count,
                         module_to_kernel_invocation_handle* handle)
                {
                    EXPECT_NE(data, nullptr);
                    EXPECT_EQ(released, nullptr);
                    EXPECT_EQ(reference_counts, nullptr);
                    EXPECT_EQ(released_pointers, nullptr);
                    EXPECT_NE(released_count, nullptr);
                    EXPECT_EQ(*released_count, 0);
                    EXPECT_NE(handle, nullptr);

                    auto *const fields = reinterpret_cast<fields_type*>(data);

                    EXPECT_EQ(fields->m_status_type, value_type_handle::invalid);
                    EXPECT_EQ(fields->m_status_data, nullptr);

                    *handle = module_to_kernel_invocation_handle(123);
                    pack_field_iterator(data)
                        .write(exception_handle::invalid)
                        .write(exception_kind::invalid)
                        .write(value_type_handle::messengers__status)
                        .write(nullptr);

                    return invocation_status::success;
                }
            );

        shared_ptr<status> argument = nullptr;
        a->truncate(argument);

        EXPECT_NE(argument, nullptr);
        EXPECT_EQ(argument->get_type(), value_type_handle::messengers__status);
    }

    {
        away_fields_type away_data {
            device_type::environment,
            600
        };

        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, invoke_synchronously(object_handle(1), interface_handle::messengers__status_handler, GET_METHOD_ID(status_handler::truncate), _, _, _, _, _, _))
            .Times(1)
            .WillOnce([&away_data](object_handle, interface_handle, method_id,
                         void* data, object_handle* released, uint64_t* reference_counts,
                         void** released_pointers, uint64_t* released_count,
                         module_to_kernel_invocation_handle* handle)
                {
                    EXPECT_NE(data, nullptr);
                    EXPECT_EQ(released, nullptr);
                    EXPECT_EQ(reference_counts, nullptr);
                    EXPECT_EQ(released_pointers, nullptr);
                    EXPECT_NE(released_count, nullptr);
                    EXPECT_EQ(*released_count, 0);
                    EXPECT_NE(handle, nullptr);

                    auto *const fields = reinterpret_cast<fields_type*>(data);

                    EXPECT_EQ(fields->m_status_type, value_type_handle::messengers__status);
                    EXPECT_EQ(fields->m_status_data, nullptr);

                    *handle = module_to_kernel_invocation_handle(123);
                    pack_field_iterator(data)
                        .write(exception_handle::invalid)
                        .write(exception_kind::invalid)
                        .write(value_type_handle::messengers__status_online_away)
                        .write(&away_data);

                    return invocation_status::success;
                }
            );
        EXPECT_CALL(kernel, release_kernel_side_containers(module_to_kernel_invocation_handle(123)));

        shared_ptr<status> argument = make_shared<status>();
        a->truncate(argument);

        EXPECT_NE(argument, nullptr);
        EXPECT_EQ(argument->get_type(), value_type_handle::messengers__status_online_away);
        EXPECT_EQ(dynamic_pointer_cast<status_online_away>(argument)->device_type, device_type::environment);
        EXPECT_EQ(dynamic_pointer_cast<status_online_away>(argument)->minutes, 600);
    }

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, release_object_handle(object_handle(1))).Times(1).WillOnce(Return(true));
        a = nullptr;
    }
}

TEST(DynamicStructure, InboundCallInputParameter)
{
    std::shared_ptr<status_handler__mock> a;

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, register_host_object(interface_handle::messengers__status_handler, 0, _))
            .WillOnce(Return(object_handle(1234)));
        a = std::make_shared<status_handler__mock>();
    }

    EXPECT_EQ(a->get_handle(), object_handle(1234));

    struct fields_type
    {
        alignas(handle_type) exception_throw_mode m_mode;
        alignas(handle_type) handle_type m_handle;
        alignas(handle_type) value_type_handle m_status_type;
        alignas(handle_type) void* m_status_data;
    };

    {
        fields_type data {
            exception_throw_mode::no_exception,
            invalid_handle,
            value_type_handle::invalid,
            nullptr
        };
        void* invocation_pointer = nullptr;

        EXPECT_CALL(*a, truncate(_)).WillOnce([](auto &status){
            EXPECT_EQ(status, nullptr);
        });
        handle_call(object_handle(1234), static_cast<object*>(a.get()), interface_handle::messengers__status_handler,
                    GET_METHOD_ID(status_handler::truncate), &data, kernel_to_module_invocation_handle(1),
                    &invocation_pointer);

        EXPECT_EQ(data.m_mode, exception_throw_mode::no_exception);
        EXPECT_EQ(data.m_handle, invalid_handle);

        EXPECT_EQ(data.m_status_type, value_type_handle::invalid);
        EXPECT_EQ(data.m_status_data, nullptr);
    }

    {
        fields_type data {
            exception_throw_mode::no_exception,
            invalid_handle,
            value_type_handle::messengers__status,
            nullptr
        };
        void* invocation_pointer = nullptr;

        EXPECT_CALL(*a, truncate(_)).WillOnce([](auto &status){
            EXPECT_NE(status, nullptr);
            EXPECT_EQ(status->get_type(), value_type_handle::messengers__status);
        });
        handle_call(object_handle(1234), static_cast<object*>(a.get()), interface_handle::messengers__status_handler,
                    GET_METHOD_ID(status_handler::truncate), &data, kernel_to_module_invocation_handle(1),
                    &invocation_pointer);

        EXPECT_EQ(data.m_mode, exception_throw_mode::no_exception);
        EXPECT_EQ(data.m_handle, invalid_handle);

        EXPECT_EQ(data.m_status_type, value_type_handle::messengers__status);
        EXPECT_EQ(data.m_status_data, nullptr);
    }

    {
        struct away_fields_type {
            alignas(uint64_t) device_type m_device_type;
            alignas(uint64_t) uint64_t m_minutes_away;
        } away_fields {
            device_type::personal_computer,
            480
        };

        fields_type data {
            exception_throw_mode::no_exception,
            invalid_handle,
            value_type_handle::messengers__status_online_away,
            &away_fields
        };
        void* invocation_pointer = nullptr;

        EXPECT_CALL(*a, truncate(_)).WillOnce([](auto &status){
            EXPECT_NE(status, nullptr);
            EXPECT_EQ(status->get_type(), value_type_handle::messengers__status_online_away);
            EXPECT_EQ(dynamic_pointer_cast<status_online_away>(status)->device_type, device_type::personal_computer);
            EXPECT_EQ(dynamic_pointer_cast<status_online_away>(status)->minutes, 480);
        });
        handle_call(object_handle(1234), static_cast<object*>(a.get()), interface_handle::messengers__status_handler,
                    GET_METHOD_ID(status_handler::truncate), &data, kernel_to_module_invocation_handle(1),
                    &invocation_pointer);

        EXPECT_EQ(data.m_mode, exception_throw_mode::no_exception);
        EXPECT_EQ(data.m_handle, invalid_handle);

        EXPECT_EQ(data.m_status_type, value_type_handle::messengers__status_online_away);
        EXPECT_EQ(data.m_status_data, &away_fields);
    }

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, release_object_handle(object_handle(1234))).Times(1).WillOnce(Return(true));
        a = nullptr;
    }
}

TEST(DynamicStructure, InboundCallParameterChanged)
{
    std::shared_ptr<status_handler__mock> a;

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, register_host_object(interface_handle::messengers__status_handler, 0, _))
            .WillOnce(Return(object_handle(1234)));
        a = std::make_shared<status_handler__mock>();
    }

    EXPECT_EQ(a->get_handle(), object_handle(1234));

    struct fields_type
    {
        alignas(handle_type) exception_throw_mode m_mode;
        alignas(handle_type) handle_type m_handle;
        alignas(handle_type) value_type_handle m_status_type;
        alignas(handle_type) void* m_status_data;
    };

    {
        fields_type data {
            exception_throw_mode::no_exception,
            invalid_handle,
            value_type_handle::invalid,
            nullptr
        };
        void* invocation_pointer = nullptr;

        EXPECT_CALL(*a, truncate(_)).WillOnce([](auto &status){
            EXPECT_EQ(status, nullptr);
            status = make_shared<status_online_away>(device_type::mobile_device, 1);
        });
        handle_call(object_handle(1234), static_cast<object*>(a.get()), interface_handle::messengers__status_handler,
                    GET_METHOD_ID(status_handler::truncate), &data, kernel_to_module_invocation_handle(1),
                    &invocation_pointer);

        EXPECT_EQ(data.m_mode, exception_throw_mode::no_exception);
        EXPECT_EQ(data.m_handle, invalid_handle);

        EXPECT_EQ(data.m_status_type, value_type_handle::messengers__status_online_away);
        EXPECT_NE(data.m_status_data, nullptr);

        struct away_fields_type {
            alignas(uint64_t) device_type m_device_type;
            alignas(uint64_t) uint64_t m_minutes_away;
        } *const away_fields = reinterpret_cast<away_fields_type*>(data.m_status_data);
        EXPECT_EQ(away_fields->m_device_type, device_type::mobile_device);
        EXPECT_EQ(away_fields->m_minutes_away, 1);

        EXPECT_NE(invocation_pointer, nullptr);
        release_inbound_call_data(kernel_to_module_invocation_handle(1), invocation_pointer);
    }

    {
        fields_type data {
            exception_throw_mode::no_exception,
            invalid_handle,
            value_type_handle::messengers__status,
            nullptr
        };
        void* invocation_pointer = nullptr;

        EXPECT_CALL(*a, truncate(_)).WillOnce([](auto &status){
            EXPECT_NE(status, nullptr);
            EXPECT_EQ(status->get_type(), value_type_handle::messengers__status);
            status = nullptr;
        });
        handle_call(object_handle(1234), static_cast<object*>(a.get()), interface_handle::messengers__status_handler,
                    GET_METHOD_ID(status_handler::truncate), &data, kernel_to_module_invocation_handle(1),
                    &invocation_pointer);

        EXPECT_EQ(data.m_mode, exception_throw_mode::no_exception);
        EXPECT_EQ(data.m_handle, invalid_handle);

        EXPECT_EQ(data.m_status_type, value_type_handle::invalid);
        EXPECT_EQ(data.m_status_data, nullptr);

        EXPECT_EQ(invocation_pointer, nullptr);
    }

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, release_object_handle(object_handle(1234))).Times(1).WillOnce(Return(true));
        a = nullptr;
    }
}
