#include <sivdk/outgoing_call.h>
#include <sivdk/module_functions.h>
#include <sivdk/call_data/proxy.h>
#include <gtest/gtest.h>
#include <sivdk-mocks/kernel_mock.h>
#include <sivdk-mocks/method_defines.h>

#include <sivtypes/graphics/graphics_2d/canvas.h>
#include <sivtypes/graphics/graphics_2d/size.h>
#include <sivtypes/physics/mechanics/body.h>
#include <sivtypes/physics/mechanics/length_value.h>
#include <sivtypes/physics/mechanics/real_world_length_unit.h>
#include <sivtypes/physics/mechanics/real_world_length_unit_info.h>
#include <sivtypes/physics/mechanics/rest_frame_length.h>
#include <sivmocks/physics/mechanics/body.h>
#include <sivmocks/graphics/graphics_2d/canvas.h>
#include <sivdk/exception_kind.h>
#include <sivdk/value_type_handle.h>

using namespace sivdk;
using namespace call_data;
using namespace std;
using namespace mocks;
using namespace testing;
using namespace sivtypes::physics::mechanics;
using namespace sivtypes::graphics::graphics_2d;


TEST(FixedStructure, OutboundCallReturnValue)
{
    object_proxy_handle_holder proxy_handle(object_handle(1));
    auto a = std::make_shared<body__proxy>(&proxy_handle);

    {
        handle_type meters;
        pack_field_iterator(&meters).write(real_world_length_unit::meter);

        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, invoke_synchronously(object_handle(1), interface_handle::physics__mechanics__body, GET_METHOD_ID(sivtypes::physics::mechanics::body::get_width), _, _, _, _, _, _))
            .Times(1)
            .WillOnce([&meters](object_handle, interface_handle, method_id,
                         void* data, object_handle* released, uint64_t* reference_counts,
                         void** released_pointers, uint64_t* released_count,
                         module_to_kernel_invocation_handle* handle)
                {
                    EXPECT_NE(data, nullptr);
                    EXPECT_EQ(released, nullptr);
                    EXPECT_EQ(reference_counts, nullptr);
                    EXPECT_EQ(released_pointers, nullptr);
                    EXPECT_NE(released_count, nullptr);
                    EXPECT_EQ(*released_count, 0);
                    EXPECT_NE(handle, nullptr);

                    *handle = module_to_kernel_invocation_handle(123);
                    pack_field_iterator(data)
                        .write(exception_handle::invalid)
                        .write(exception_kind::invalid)
                        .write(static_cast<double>(2.58))
                        .write(value_type_handle::physics__mechanics__real_world_length_unit_info)
                        .write(&meters)
                        .write(value_type_handle::physics__mechanics__rest_frame_length)
                        .write(nullptr);

                    return invocation_status::success;
                }
            );
        EXPECT_CALL(kernel, release_kernel_side_containers(module_to_kernel_invocation_handle(123)));
        auto width = a->get_width();

        EXPECT_EQ(width.value, 2.58);

        auto unit_info = dynamic_pointer_cast<real_world_length_unit_info>(width.unit);
        ASSERT_NE(unit_info, nullptr);
        EXPECT_EQ(unit_info->unit, real_world_length_unit::meter);

        auto unit_conversion_context = dynamic_pointer_cast<rest_frame_length>(width.unit_conversion_context);
        ASSERT_NE(unit_conversion_context, nullptr);
    }

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, invoke_synchronously(object_handle(1), interface_handle::physics__mechanics__body, GET_METHOD_ID(sivtypes::physics::mechanics::body::get_width), _, _, _, _, _, _))
            .Times(1)
            .WillOnce([](object_handle, interface_handle, method_id,
                         void* data, object_handle* released, uint64_t* reference_counts,
                         void** released_pointers, uint64_t* released_count,
                         module_to_kernel_invocation_handle* handle)
                {
                    EXPECT_NE(data, nullptr);
                    EXPECT_EQ(released, nullptr);
                    EXPECT_EQ(reference_counts, nullptr);
                    EXPECT_EQ(released_pointers, nullptr);
                    EXPECT_NE(released_count, nullptr);
                    EXPECT_EQ(*released_count, 0);
                    EXPECT_NE(handle, nullptr);

                    *handle = module_to_kernel_invocation_handle(123);
                    pack_field_iterator(data)
                        .write(exception_handle::invalid)
                        .write(exception_kind::invalid)
                        .write(static_cast<double>(2.58))
                        .write(value_type_handle::invalid)
                        .write(nullptr)
                        .write(value_type_handle::invalid)
                        .write(nullptr);

                    return invocation_status::success;
                }
            );
        auto width = a->get_width();

        EXPECT_EQ(width.value, 2.58);
        EXPECT_EQ(width.unit, nullptr);
        EXPECT_EQ(width.unit_conversion_context, nullptr);
    }

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, release_object_handle(object_handle(1))).Times(1).WillOnce(Return(true));
        a = nullptr;
    }
}

TEST(FixedStructure, InboundCallReturnValue)
{
    std::shared_ptr<body__mock> a;

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, register_host_object(interface_handle::physics__mechanics__body, 0, _))
            .WillOnce(Return(object_handle(1234)));
        a = std::make_shared<body__mock>();
    }

    EXPECT_EQ(a->get_handle(), object_handle(1234));

    {
        struct unit_info
        {
            real_world_length_unit m_unit;
        };

        struct
        {
            alignas(handle_type) exception_throw_mode m_mode;
            alignas(handle_type) handle_type m_handle;
            alignas(handle_type) double m_value;
            alignas(handle_type) value_type_handle m_unit_type;
            alignas(handle_type) unit_info* m_unit;
            alignas(handle_type) value_type_handle m_unit_conversion_type;
            alignas(handle_type) void* m_unit_conversion;
        } data {
            exception_throw_mode::no_exception,
            invalid_handle,
            0.0,
            value_type_handle::invalid,
            nullptr,
            value_type_handle::invalid,
            nullptr
        };
        void* invocation_pointer = nullptr;

        length_value result{
            2.75,
            make_shared<real_world_length_unit_info>(real_world_length_unit::inch),
            make_shared<rest_frame_length>()
        };
        EXPECT_CALL(*a, get_width()).WillOnce(Return(result));
        handle_call(object_handle(1234), static_cast<object*>(a.get()), interface_handle::physics__mechanics__body,
                    GET_METHOD_ID(sivtypes::physics::mechanics::body::get_width), &data, kernel_to_module_invocation_handle(1),
                    &invocation_pointer);

        EXPECT_EQ(data.m_mode, exception_throw_mode::no_exception);
        EXPECT_EQ(data.m_handle, invalid_handle);

        EXPECT_EQ(data.m_value, 2.75);

        EXPECT_EQ(data.m_unit_type, value_type_handle::physics__mechanics__real_world_length_unit_info);
        EXPECT_NE(data.m_unit, nullptr);
        EXPECT_EQ(data.m_unit->m_unit, real_world_length_unit::inch);

        EXPECT_EQ(data.m_unit_conversion_type, value_type_handle::physics__mechanics__rest_frame_length);
        EXPECT_EQ(data.m_unit_conversion, nullptr);

        EXPECT_NE(invocation_pointer, nullptr);
        release_inbound_call_data(kernel_to_module_invocation_handle(1), invocation_pointer);
    }

    {
        struct unit_info
        {
            real_world_length_unit m_unit;
        };

        struct
        {
            alignas(handle_type) exception_throw_mode m_mode;
            alignas(handle_type) handle_type m_handle;
            alignas(handle_type) double m_value;
            alignas(handle_type) value_type_handle m_unit_type;
            alignas(handle_type) unit_info* m_unit;
            alignas(handle_type) value_type_handle m_unit_conversion_type;
            alignas(handle_type) void* m_unit_conversion;
        } data {
            exception_throw_mode::no_exception,
            invalid_handle,
            0.0,
            value_type_handle::invalid,
            nullptr,
            value_type_handle::invalid,
            nullptr
        };
        void* invocation_pointer = nullptr;

        length_value result{
            10.01,
            nullptr,
            nullptr
        };
        EXPECT_CALL(*a, get_width()).WillOnce(Return(result));
        handle_call(object_handle(1234), static_cast<object*>(a.get()), interface_handle::physics__mechanics__body,
                    GET_METHOD_ID(sivtypes::physics::mechanics::body::get_width), &data, kernel_to_module_invocation_handle(1),
                    &invocation_pointer);

        EXPECT_EQ(data.m_mode, exception_throw_mode::no_exception);
        EXPECT_EQ(data.m_handle, invalid_handle);

        EXPECT_EQ(data.m_value, 10.01);

        EXPECT_EQ(data.m_unit_type, value_type_handle::invalid);
        EXPECT_EQ(data.m_unit, nullptr);

        EXPECT_EQ(data.m_unit_conversion_type, value_type_handle::invalid);
        EXPECT_EQ(data.m_unit_conversion, nullptr);

        EXPECT_EQ(invocation_pointer, nullptr);
    }

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, release_object_handle(object_handle(1234))).Times(1).WillOnce(Return(true));
        a = nullptr;
    }
}

TEST(FixedStructure, OutboundCallInputParameter)
{
    object_proxy_handle_holder proxy_handle(object_handle(1));
    auto a = std::make_shared<canvas__proxy>(&proxy_handle);

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, invoke_synchronously(object_handle(1), interface_handle::graphics__graphics_2d__canvas, GET_METHOD_ID(canvas::fit_size), _, _, _, _, _, _))
            .Times(1)
            .WillOnce([](object_handle, interface_handle, method_id,
                         void* data, object_handle* released, uint64_t* reference_counts,
                         void** released_pointers, uint64_t* released_count,
                         module_to_kernel_invocation_handle* handle)
                {
                    EXPECT_NE(data, nullptr);
                    EXPECT_EQ(released, nullptr);
                    EXPECT_EQ(reference_counts, nullptr);
                    EXPECT_EQ(released_pointers, nullptr);
                    EXPECT_NE(released_count, nullptr);
                    EXPECT_EQ(*released_count, 0);
                    EXPECT_NE(handle, nullptr);

                    uint64_t width = 0;
                    uint64_t height = 0;

                    pack_field_iterator(data)
                        .skip()
                        .skip()
                        .read(&width)
                        .read(&height);

                    EXPECT_EQ(width, 640);
                    EXPECT_EQ(height, 480);

                    *handle = module_to_kernel_invocation_handle(123);
                    pack_field_iterator(data)
                        .write(exception_handle::invalid)
                        .write(exception_kind::invalid);

                    return invocation_status::success;
                }
            );

        sivtypes::graphics::graphics_2d::size argument(640, 480);
        a->fit_size(argument);

        EXPECT_EQ(argument.width, 640);
        EXPECT_EQ(argument.height, 480);
    }

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, release_object_handle(object_handle(1))).Times(1).WillOnce(Return(true));
        a = nullptr;
    }
}

TEST(FixedStructure, OutboundCallParameterChange)
{
    object_proxy_handle_holder proxy_handle(object_handle(1));
    auto a = std::make_shared<canvas__proxy>(&proxy_handle);

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, invoke_synchronously(object_handle(1), interface_handle::graphics__graphics_2d__canvas, GET_METHOD_ID(canvas::fit_size), _, _, _, _, _, _))
            .Times(1)
            .WillOnce([](object_handle, interface_handle, method_id,
                         void* data, object_handle* released, uint64_t* reference_counts,
                         void** released_pointers, uint64_t* released_count,
                         module_to_kernel_invocation_handle* handle)
                {
                    EXPECT_NE(data, nullptr);
                    EXPECT_EQ(released, nullptr);
                    EXPECT_EQ(reference_counts, nullptr);
                    EXPECT_EQ(released_pointers, nullptr);
                    EXPECT_NE(released_count, nullptr);
                    EXPECT_EQ(*released_count, 0);
                    EXPECT_NE(handle, nullptr);

                    uint64_t width = 0;
                    uint64_t height = 0;

                    pack_field_iterator(data)
                        .skip()
                        .skip()
                        .read(&width)
                        .read(&height);

                    EXPECT_EQ(width, 640);
                    EXPECT_EQ(height, 480);

                    *handle = module_to_kernel_invocation_handle(123);
                    pack_field_iterator(data)
                        .write(exception_handle::invalid)
                        .write(exception_kind::invalid)
                        .write(uint64_t(1024))
                        .write(uint64_t(768));

                    return invocation_status::success;
                }
            );

        sivtypes::graphics::graphics_2d::size argument(640, 480);
        a->fit_size(argument);

        EXPECT_EQ(argument.width, 1024);
        EXPECT_EQ(argument.height, 768);
    }

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, release_object_handle(object_handle(1))).Times(1).WillOnce(Return(true));
        a = nullptr;
    }
}

TEST(FixedStructure, InboundCallInputParameter)
{
    std::shared_ptr<canvas__mock> a;

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, register_host_object(interface_handle::graphics__graphics_2d__canvas, 0, _))
            .WillOnce(Return(object_handle(1234)));
        a = std::make_shared<canvas__mock>();
    }

    EXPECT_EQ(a->get_handle(), object_handle(1234));

    {
        struct
        {
            alignas(handle_type) exception_throw_mode m_mode;
            alignas(handle_type) handle_type m_handle;
            alignas(handle_type) uint64_t m_width;
            alignas(handle_type) uint64_t m_height;
        } data {
            exception_throw_mode::no_exception,
            invalid_handle,
            640,
            480
        };
        void* invocation_pointer = nullptr;

        EXPECT_CALL(*a, fit_size(_)).WillOnce([](sivtypes::graphics::graphics_2d::size& size) {
            EXPECT_EQ(size.width, 640);
            EXPECT_EQ(size.height, 480);
        });
        handle_call(object_handle(1234), static_cast<object*>(a.get()), interface_handle::graphics__graphics_2d__canvas,
                    GET_METHOD_ID(canvas::fit_size), &data, kernel_to_module_invocation_handle(1),
                    &invocation_pointer);

        EXPECT_EQ(data.m_mode, exception_throw_mode::no_exception);
        EXPECT_EQ(data.m_handle, invalid_handle);

        EXPECT_EQ(data.m_width, 640);
        EXPECT_EQ(data.m_height, 480);
    }

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, release_object_handle(object_handle(1234))).Times(1).WillOnce(Return(true));
        a = nullptr;
    }
}

TEST(FixedStructure, InboundCallParameterChanged)
{
    std::shared_ptr<canvas__mock> a;

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, register_host_object(interface_handle::graphics__graphics_2d__canvas, 0, _))
            .WillOnce(Return(object_handle(1234)));
        a = std::make_shared<canvas__mock>();
    }

    EXPECT_EQ(a->get_handle(), object_handle(1234));

    {
        struct
        {
            alignas(handle_type) exception_throw_mode m_mode;
            alignas(handle_type) handle_type m_handle;
            alignas(handle_type) uint64_t m_width;
            alignas(handle_type) uint64_t m_height;
        } data {
            exception_throw_mode::no_exception,
            invalid_handle,
            640,
            480
        };
        void* invocation_pointer = nullptr;

        EXPECT_CALL(*a, fit_size(_)).WillOnce([](sivtypes::graphics::graphics_2d::size& size) {
            EXPECT_EQ(size.width, 640);
            EXPECT_EQ(size.height, 480);

            size.width=800;
            size.height=600;
        });
        handle_call(object_handle(1234), static_cast<object*>(a.get()), interface_handle::graphics__graphics_2d__canvas,
                    GET_METHOD_ID(canvas::fit_size), &data, kernel_to_module_invocation_handle(1),
                    &invocation_pointer);

        EXPECT_EQ(data.m_mode, exception_throw_mode::no_exception);
        EXPECT_EQ(data.m_handle, invalid_handle);

        EXPECT_EQ(data.m_width, 800);
        EXPECT_EQ(data.m_height, 600);
    }

    {
        testing::StrictMock<kernel_mock> kernel;
        EXPECT_CALL(kernel, release_object_handle(object_handle(1234))).Times(1).WillOnce(Return(true));
        a = nullptr;
    }
}
